import torch
import torch.nn as nn


class EnsembleFC(nn.Module):
    def __init__(self, in_features: int, out_features: int, ensemble_size: int, weight_decay: float = 0., bias: bool = True) -> None:
        super(EnsembleFC, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.ensemble_size = ensemble_size
        self.weight = nn.Parameter(torch.Tensor(ensemble_size, in_features, out_features))
        torch.nn.init.xavier_uniform_(self.weight)
        #torch.nn.init.trunc_normal_(self.weight)
        self.weight_decay = weight_decay
        if bias:
            self.bias = nn.Parameter(torch.zeros(ensemble_size, out_features))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        pass

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        w_times_x = torch.bmm(input, self.weight)
        return torch.add(w_times_x, self.bias[:, None, :])  # w times x + b

    def extra_repr(self) -> str:
        return 'in_features={}, out_features={}, bias={}'.format(
            self.in_features, self.out_features, self.bias is not None
        )



from pytorchrl.envs.minigrid.custom_environments.deceiving_rewards import DeceivingRewardsEnv
from pytorchrl.envs.minigrid.custom_environments.multiple_deceiving_rewards import MultipleDeceivingRewardsEnv


from gymnasium.envs.registration import register

from minigrid import minigrid_env, wrappers
from minigrid.core import roomgrid
from minigrid.core.world_object import Wall


def register_custom_minigrid_envs():

    # DeceivingRewards
    # ----------------

    register(
        id="MiniGrid-DeceivingRewards-v0",
        entry_point="pytorchrl.envs.minigrid.custom_environments:DeceivingRewardsEnv",
    )

    # MultipleDeceivingRewards
    # ----------------

    register(
        id="MiniGrid-MultipleDeceivingRewards-v0",
        entry_point="pytorchrl.envs.minigrid.custom_environments:MultipleDeceivingRewardsEnv",
    )





import gym
from gym.spaces.box import Box


class TransposeImagesIfRequired(gym.ObservationWrapper):
    """
    When environment observations are images, this wrapper transposes
    the axis. It is useful when the images have shape (W,H,C), as they can be
    transposed "on the fly" to (C,W,H) for PyTorch convolutions to be applied.

    Parameters
    ----------
    env : gym.Env
        Original Gym environment, previous to applying the wrapper.
    op : list
        New axis ordering.
    """

    def __init__(self, env=None, op=[2, 0, 1]):
        """Transpose observation space for images"""
        super(TransposeImagesIfRequired, self).__init__(env)

        self.op = op

        if isinstance(self.observation_space, gym.spaces.Box) and \
                len(self.observation_space.shape) == 3:
            obs_shape = self.observation_space.shape
            self.observation_space = Box(
                self.observation_space.low[0, 0, 0],
                self.observation_space.high[0, 0, 0],
                [obs_shape[self.op[0]], obs_shape[self.op[1]], obs_shape[self.op[2]]],
                dtype=self.observation_space.dtype)

        elif isinstance(self.observation_space, gym.spaces.Dict):
            for k in self.observation_space.spaces:
                if isinstance(self.observation_space[k], gym.spaces.Box) and \
                        len(self.observation_space[k].shape) == 3:
                    obs_shape = self.observation_space[k].shape
                    self.observation_space[k] = Box(
                        self.observation_space[k].low[0, 0, 0],
                        self.observation_space[k].high[0, 0, 0],
                        [obs_shape[self.op[0]], obs_shape[self.op[1]], obs_shape[self.op[2]]],
                        dtype=self.observation_space.dtype)

    def observation(self, ob):
        """Transpose observation"""

        if isinstance(ob, dict):
            for k in ob:
                if len(ob[k].shape) == 3:
                    ob[k] = ob[k].transpose(self.op[0], self.op[1], self.op[2])
        else:
            if len(ob.shape) == 3:
                ob = ob.transpose(self.op[0], self.op[1], self.op[2])

        return ob



import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim

import pytorchrl as prl
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients


class MPC_RS(Algorithm):
    """
    Model-Based MPC Random Shooting (RS) class.
    Trains a model of the environment and uses RS to select actions.

    Parameters
    ----------
    lr: float
        Dynamics model learning rate.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        actor class instance.
    device : torch.device
        CPU or specific GPU where class computations will take place.
    mb_epochs : int
        Training epochs for the dynamics model.
    start_steps: int
        Number of steps collected with initial random policy.
    update_every : int
         Amount of data collected in between dynamics model updates.
    action_noise :
        Exploration noise.
    mini_batch_size : int
        Size of actor update batches.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations.
    """

    def __init__(self,
                 lr,
                 envs,
                 actor,
                 device,
                 mb_epochs,
                 start_steps,
                 update_every,
                 action_noise,
                 max_grad_norm,
                 mini_batch_size,
                 num_test_episodes,
                 test_every):

        # ---- General algo attributes ----------------------------------------

        # Number of steps collected with initial random policy
        self._start_steps = int(start_steps)

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = int(mb_epochs)

        # Tracks the number of times data is reused
        self.mb_train_epochs = 0

        # Number of data samples collected between network update stages
        self._update_every = int(update_every)

        # Number mini batches per epoch
        self._num_mini_batch = int(1)  # Depends on how much data is available

        # Size of update mini batches
        self._mini_batch_size = int(mini_batch_size)

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = num_test_episodes

        # ---- RS-specific attributes ----------------------------------------

        # Number of episodes to complete when testing
        self.iter = 0
        self.envs = envs
        self.actor = actor
        self.device = device
        self.reuse_data = False
        self.action_noise = action_noise
        self.max_grad_norm = max_grad_norm

        if self.actor.action_type == "discrete":
            self.get_rollout_actions = self._get_discrete_actions
        elif self.actor.action_type == "continuous":
            self.get_rollout_actions = self._get_continuous_actions
        else:
            raise ValueError("Selected action type does not exist!")

        # ----- Optimizers ----------------------------------------------------

        self.dynamics_optimizer = optim.Adam(self.actor.dynamics_model.parameters(), lr=lr)
        self.loss_func = torch.nn.MSELoss()

    @classmethod
    def create_factory(
            cls,
            lr,
            start_steps,
            update_every,
            mb_epochs,
            action_noise,
            mini_batch_size,
            test_every=10,
            max_grad_norm=0.5,
            num_test_episodes=3):
        """
        Returns a function to create a new Model-Based MPC instance.

        Parameters
        ----------
        lr: float
            Dynamics model learning rate.
        start_steps: int
            Number of steps collected with initial random policy.
        update_every : int
             Amount of data collected in between dynamics model updates.
        mb_epochs : int
            Training epochs for the dynamics model.
        action_noise :
            Exploration noise.
        mini_batch_size : int
            Size of actor update batches.
        test_every : int
            Regularity of test evaluations.
        num_test_episodes : int
            Number of episodes to complete in each test phase.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new MPC_RS class instance.
        algo_name : str
            Name of the algorithm.
        """

        def create_algo_instance(device, actor, envs):
            return cls(lr=lr,
                       envs=envs,
                       actor=actor,
                       device=device,
                       mb_epochs=update_every,
                       start_steps=start_steps,
                       update_every=update_every,
                       action_noise=action_noise,
                       max_grad_norm=max_grad_norm,
                       mini_batch_size=mini_batch_size,
                       num_test_episodes=num_test_episodes,
                       test_every=test_every)

        return create_algo_instance, prl.MPC_RS

    @property
    def gamma(self):
        """Returns discount factor gamma."""
        return None

    @property
    def start_steps(self):
        """Returns the number of steps to collect with initial random policy."""
        return self._start_steps

    @property
    def num_epochs(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_epochs

    @property
    def update_every(self):
        """
        Returns the number of data samples collected between
        network update stages.
        """
        return self._update_every

    @property
    def num_mini_batch(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_mini_batch

    @property
    def mini_batch_size(self):
        """
        Returns the number of mini batches per epoch.
        """
        return self._mini_batch_size

    @property
    def test_every(self):
        """Number of network updates between test evaluations."""
        return self._test_every

    @property
    def num_test_episodes(self):
        """
        Returns the number of episodes to complete when testing.
        """
        return self._num_test_episodes

    def _get_discrete_actions(self, ) -> torch.Tensor:
        """Samples random discrete actions"""
        return torch.randint(self.actor.action_dims, size=(
            self.actor.n_planner, self.actor.horizon, 1)).to(self.device)

    def _get_continuous_actions(self, ) -> torch.Tensor:
        """Samples random continuous actions"""
        actions = np.random.uniform(
            low=self.actor.action_low,
            high=self.actor.action_high,
            size=(self.actor.n_planner, self.actor.horizon, self.actor.action_dims))
        return torch.from_numpy(actions).to(self.device).float()

    def compute_returns(self, states: torch.Tensor, actions: torch.Tensor, model: torch.nn.Module):
        """
        Calculates the trajectory returns

        Parameters
        ----------
        states: torch.Tensor
            Trajectory states
        actions: torch.Tensor
            Trajectory actions
        model: dynamics Model
            Calculates the next states and rewards

        Returns
        -------
        returns: torch.Tensor
            Trajectory returns of the RS MPC

        """
        returns = torch.zeros((self.actor.n_planner, 1)).to(self.device)
        for t in range(self.actor.horizon):
            with torch.no_grad():
                states, rewards = model.predict(states, actions[:, t, :])
            returns += rewards

        return returns

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        Does the MPC search with random shooting action planning process.

        Parameters
        ----------
        obs: torch.tensor
            Current world observation
        rhs: dict
            RNN recurrent hidden states.
        done: torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic: bool
            Whether to randomly sample action from predicted distribution or taking the mode.

        Returns
        -------
        action: torch.tensor
            Predicted next action.
        clipped_action: torch.tensor
            Predicted next action (clipped to be within action space).
        rhs: batch
            Actor recurrent hidden state.
        other: dict
            Additional MPC predictions, which are not used in other algorithms.
        """

        with torch.no_grad():

            initial_states = obs.repeat((self.actor.n_planner, 1)).to(self.device)
            rollout_actions = self.get_rollout_actions()
            returns = self.compute_returns(initial_states, rollout_actions, self.actor.dynamics_model)
            action = rollout_actions[:, 0, :][returns.argmax()]

            if self.action_noise and self.actor.action_type == "continuous":
                action += torch.normal(
                    torch.zeros(action.shape),
                    torch.ones(action.shape) * 0.005).to(self.device)

            clipped_action = action

        if self.actor.dynamics_model.unscale:
            action = self.actor.dynamics_model.unscale(action)
            clipped_action = self.actor.dynamics_model.unscale(clipped_action)

        return action.unsqueeze(0), clipped_action.unsqueeze(0), rhs, {}

    def training_step(self, batch):
        """Does the forward pass and loss calculation of the dynamics model given the training data.

        Parameters
        ----------
        batch: dict
            Training data with inputs and labels

        Returns
        -------
            torch.Tensor: Returns the training loss
        """
        train_inputs = batch["train_input"]
        train_labels = batch["train_label"]

        self.actor.train()
        prediction = self.actor.dynamics_model.model(train_inputs)
        loss = self.loss_func(prediction, train_labels)
        return loss

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        batch: dict
            data batch containing all required tensors to compute dynamics model losses.
        grads_to_cpu: bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads: list of tensors
            List of actor_critic gradients.
        info: dict
            Dict containing current dynamics model iteration information.
        """
        if batch["batch_number"] == 0:

            # TODO: add reinitialization
            # reinitializes model for new training
            # if self.iter != 0 and self.mb_train_epochs == 0:
            #     self.actor.reinitialize_dynamics_model()
            #     self.actor.to(self.device)
            #     self.dynamics_optimizer = optim.Adam(self.actor.dynamics_model.parameters(), lr=self.lr)

            self.reuse_data = True
            self.mb_train_epochs += 1

        if self.mb_train_epochs == self.num_epochs:
            self.reuse_data = False
            self.mb_train_epochs = 0

        train_loss = self.training_step(batch)

        self.dynamics_optimizer.zero_grad()
        train_loss.backward()
        nn.utils.clip_grad_norm_(self.actor.dynamics_model.parameters(), self.max_grad_norm)
        dyna_grads = get_gradients(self.actor.dynamics_model, grads_to_cpu=grads_to_cpu)

        info = {"train_loss": train_loss.item()}
        grads = {"dyna_grads": dyna_grads}

        # once break condition is used set reuse_data to False
        return grads, info

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.
        Update also target networks.

        Parameters
        ----------
        gradients : list of tensors
            List of actor gradients.
        """
        if gradients is not None:
            set_gradients(
                self.actor.dynamics_model,
                gradients=gradients["dyna_grads"], device=self.device)

        self.dynamics_optimizer.step()
        self.iter += 1

    def set_weights(self, actor_weights):
        """
        Update actor with the given weights. Update also target networks.

        Parameters
        ----------
        actor_weights : dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(actor_weights)
        self.iter += 1

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.dynamics_optimizer.param_groups:
                param_group['lr'] = new_parameter_value



import torch
import numpy as np

import pytorchrl as prl
from pytorchrl.utils import RunningMeanStd
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn


class RewardPredictor(PolicyLossAddOn):

    def __init__(self, predictor_net_factory, predictor_net_kwargs, masked_sparse_obs_ratio=0.0):
        """
        Class to train a reward predictor in parallel with the RL actor.

        Parameters
        ----------
        predictor_net_factory : func
            Method to create the network that predict the reward (output has shape (1,)).
        predictor_net_factory : dict
            Keyword argument to be used in predictor_net_factory.
        masked_sparse_obs_ratio : func
            Ratio, between 0.0 and 1.0, of states with reward 0.0 to be masked in the loss function.
        """
        self.predictor_net_kwargs = predictor_net_kwargs
        self.predictor_net_factory = predictor_net_factory
        self.masked_sparse_obs_ratio = np.clip(masked_sparse_obs_ratio, 0.0, 1.0)

    def setup(self, actor, device):
        """Setup addon module by creating and instance of the network."""

        self.actor = actor
        self.device = device

        # Cast behavior weights to torch tensors
        self.actor.reward_predictor = self.predictor_net_factory(**self.predictor_net_kwargs).to(device)
        self.max_pred_errors_rms = RunningMeanStd(shape=(1,), device=device)
        self.mean_pred_errors_rms = RunningMeanStd(shape=(1,), device=device)
        self.min_pred_errors_rms = RunningMeanStd(shape=(1,), device=device)
        self.actor.error_threshold = torch.nn.parameter.Parameter(
            data=torch.tensor(1000000, dtype=torch.float32), requires_grad=False)

    def compute_loss_term(self, data, actor_dist, info):
        """
        Calculate and add KL Attraction loss term.

        Parameters
        ----------
        actor_dist : torch.distributions.Distribution
            Actor action distribution for actions in data[prl.OBS]
        data : dict
            data batch containing all required tensors to compute loss term.
        info : dict
            Dictionary to store log information.

        Returns
        -------
        attraction_kl_loss_term : torch.tensor
            KL loss term.
        info : dict
            Updated info dict.
        """

        o, rhs, r = data[prl.OBS], data[prl.RHS], data[prl.REW]
        pred_r = self.actor.reward_predictor(o)
        error = torch.abs(r - pred_r)
        loss = 0.5 * error.pow(2)
        mask = torch.rand(loss.size(), device=self.device)
        mask = (mask >= self.masked_sparse_obs_ratio).float()
        mask[r != 0.0] = 1.0
        loss = (mask * loss).sum() / mask.sum()

        if len(error[r != 0.0]) > 0:
            self.max_pred_errors_rms.update(error[r != 0.0].max().reshape(-1, 1))
            self.mean_pred_errors_rms.update(error[r != 0.0].mean().reshape(-1, 1))
            self.min_pred_errors_rms.update(error[r != 0.0].min().reshape(-1, 1))
        self.actor.error_threshold.data = self.max_pred_errors_rms.mean[0].float()

        info.update({
            "reward_predictor_loss": loss.item(),
            "reward_predictor_error": error.mean().item(),
            "max_reward_pred_error_rms": self.max_pred_errors_rms.mean.float().item(),
            "mean_reward_pred_error_rms": self.mean_pred_errors_rms.mean.float().item(),
            "min_reward_pred_error_rms": self.min_pred_errors_rms.mean.float().item(),
        })

        return loss, info



import os
import time
import logging
from functools import partial
from collections import defaultdict, deque
import pytorchrl as prl

logger = logging.getLogger(__name__)


class Learner:
    """
    Task learner class. Class to manage the training process. It pushes forward the training
    process by calling the update workers and tracks progress.

    Parameters
    ----------
    scheme : Scheme
        Training scheme class instance, handling coordination of workers.
    target_steps : int
        Number of environment steps to reach to complete training.
    log_dir : str
        Target directory for model checkpoints and logs.
    """

    def __init__(self, scheme, target_steps, log_dir=None):

        # Input attributes
        self.log_dir = log_dir
        self.target_steps = target_steps
        self.update_worker = scheme.update_worker()

        # Counters and metrics
        self.num_samples_collected = 0
        self.metrics = {k: defaultdict(partial(deque, maxlen=1)) for k in prl.INFO_KEYS}

        # Record starting time
        self.start = time.time()

        logger.warning("Created Learner.")

    def step(self):
        """Takes a logical synchronous optimization step."""

        # Update step
        info = self.update_worker.step()

        actor_version = info[prl.VERSION]

        # grad_update_lag
        info[prl.SCHEME] = {}
        info[prl.SCHEME][prl.FPS] = int(self.num_samples_collected / (time.time() - self.start))
        info[prl.SCHEME][prl.PL] = actor_version[prl.GRADIENT] - actor_version[prl.COLLECTION]
        info[prl.SCHEME][prl.GA] = actor_version[prl.UPDATE] - actor_version[prl.GRADIENT]

        # Update counters
        self.num_samples_collected += info.pop(prl.NUMSAMPLES)

        # Update and log metrics
        for k in prl.INFO_KEYS:
            if k in info and isinstance(info[k], dict):
                for x, y in info[k].items():
                    if isinstance(y, (float, int)):
                        self.metrics[k][x].append(y)

    def done(self):
        """
        Return True if training has finished (target_steps reached).

        Returns
        -------
        flag : bool
            True if training has reached the target number of steps.
        """
        flag = self.num_samples_collected >= self.target_steps
        if flag:
            self.update_worker.stop()
            print("\nTraining finished!")
            time.sleep(1)
        return flag

    def get_metrics(self, add_algo_metrics=True, add_episodes_metrics=False, add_scheme_metrics=False, add_time_metrics=False):
        """Returns current value of tracked metrics."""

        m = {}

        def include_info(key):
            for k, v in self.metrics[key].items():
                m[os.path.join(key, k)] = sum(v) / len(v)

        if add_algo_metrics:
            include_info(prl.ALGORITHM)

        if add_episodes_metrics:
            include_info(prl.EPISODES)

        if add_scheme_metrics:
            include_info(prl.SCHEME)

        if add_time_metrics:
            include_info(prl.TIME)

        return m

    def print_info(self, add_algo_info=True, add_episodes_info=True, add_scheme_info=False, add_time_info=False):
        """Print relevant information about the training process"""

        def write_info(msg, key):
            msg += "\n  {}: ".format(key)
            for k, v in self.metrics[key].items():
                msg += "{} {:.4f}, ".format(k, sum(v) / len(v))
            msg = msg[:-2]
            return msg

        s = "Update {}".format(self.update_worker.actor_version)
        s += ", num samples collected {}, FPS {}".format(self.num_samples_collected,
            int(self.num_samples_collected / (time.time() - self.start)))

        if add_algo_info:
            s = write_info(s, prl.ALGORITHM)

        if add_episodes_info:
            s = write_info(s, prl.EPISODES)

        if add_scheme_info:
            s = write_info(s, prl.SCHEME)

        if add_time_info:
            s = write_info(s, prl.TIME)

        print(s, flush=True)

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm used for training,
        change its value to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        self.update_worker.update_algorithm_parameter(parameter_name, new_parameter_value)

    def save_model(self, fname="model.state_dict"):
        """
        Save currently learned actor_critic version.

        Returns
        -------
        save_name : str
            Path to saved file.
        """
        fname = os.path.join(self.log_dir, fname)
        save_name = self.update_worker.save_model(fname)
        return save_name

    def stop(self):
        """Stop all threads."""
        self.update_worker.stop()



import gym
import numpy as np
from gym import spaces
from collections import defaultdict, deque


class GenChemEnv(gym.Env):
    """Custom Environment for Generative Chemistry RL."""

    metadata = {'render.modes': ['human']}

    def __init__(self, scoring_function, vocabulary, max_length=200):
        super(GenChemEnv, self).__init__()

        self.num_episodes = 0
        self.vocabulary = vocabulary
        self.max_length = max_length
        self.scoring_function = scoring_function
        self.running_mean_valid_smiles = deque(maxlen=100)

        # Define action and observation space
        self.current_episode_length = 0
        self.action_space = gym.spaces.Discrete(len(self.vocabulary))
        self.observation_space = gym.spaces.Discrete(len(self.vocabulary))
        self.action_space.shape = (max_length,)
        self.observation_space.shape = (max_length,)
        self.current_molecule_np = -1 * np.ones(max_length)

    def step(self, action):
        """Execute one time step within the environment"""

        info = {}
        self.current_episode_length += 1
        action = "$" if self.current_episode_length == self.max_length - 1 else \
            self.vocabulary.decode_token(action)
        self.current_molecule_str += action

        if action != "$":  # If character is not $, return 0.0 reward
            reward = 0.0
            done = False

        else:  # if action is $, evaluate molecule

            # Compute score
            score = self.scoring_function(self.vocabulary.remove_start_and_end_tokens(self.current_molecule_str))

            # Sanity check
            assert isinstance(score, dict), "scoring_function has to return a dict"
            assert "score" in score.keys() or "reward" in score.keys(), \
                "scoring_function outputs requires at lest the keyword ´score´ or ´reward´"

            # Get reward
            reward = score["reward"] if "reward" in score.keys() else score["score"]

            # If score contain field "Valid", update counter
            if "valid_smile" in score.keys():
                valid = score["valid_smile"]
                self.running_mean_valid_smiles.append(1.0) if valid else \
                    self.running_mean_valid_smiles.append(0.0)

            # Update molecule
            info.update({"molecule": self.vocabulary.remove_start_and_end_tokens(self.current_molecule_str)})

            # Update valid smiles tracker
            info.update({"valid_smile": float((sum(self.running_mean_valid_smiles) / len(
                self.running_mean_valid_smiles)) * 100 if len(self.running_mean_valid_smiles) != 0.0 else 0.0)})

            # Update info with remaining values
            info.update(score)
            done = True

        new_obs = self.vocabulary.encode_token(action)
        self.current_molecule_np[self.current_episode_length] = new_obs
        return self.current_molecule_np, reward, done, info

    def reset(self):
        """
        Reset the state of the environment to an initial state.
        Return padded base molecule to match length `obs_length`.
        """
        self.num_episodes += 1
        self.current_molecule_str = "^"
        self.current_episode_length = 0
        self.current_molecule_np = -1 * np.ones(self.max_length)
        self.current_molecule_np[self.current_episode_length] = self.vocabulary.encode_token("^")
        return self.current_molecule_np

    def render(self, mode='human'):
        """Render the environment to the screen"""

        print(f'Current Molecule: {self.current_molecule}')
        print(f'Vocabulary: {self.vocabulary._tokens}')



import torch
from pytorchrl.agent.env.make_env import make_env
from pytorchrl.agent.env.vector_wrappers import VecPyTorch
from pytorchrl.agent.env.openai_baselines_dependencies.vec_envs.dummy_vec_env import DummyVecEnv
from pytorchrl.agent.env.openai_baselines_dependencies.vec_envs.subproc_vec_env import SubprocVecEnv


class VecEnv:
    """Class to handle creation of environment vectors"""

    @classmethod
    def create_factory(cls, env_fn, env_kwargs={}, vec_env_size=1, log_dir=None, info_keywords=()):
        """
        Returns a function to create a vector of environments of size
        num_processes, so it can be executed by any worker, remote or not.

        Parameters
        ----------
        env_fn : func
            Function to create the environment.
        env_kwargs : dict
            keyword arguments of env_fn.
        vec_env_size : int
            size of the vector of environments.
        log_dir : str
            Target path for envs to log information through bench.Monitor class.
        info_keywords : tuple
            Information keywords to be logged stored by bench.Monitor class.

        Returns
        -------
        make_vec_env : func
            Function to create a vector of environments.
        dummy_env.action_space : gym.Space
            Environments action space.
        dummy_env.observation_space: gym.Space
            Environments observation space.
        """

        def make_vec_env(device=torch.device("cpu"), index_col_worker=1, index_grad_worker=1, mode="train"):
            """Create and return a vector environment"""

            if mode == "train":
                env_indexes = range(1, vec_env_size + 1)
            else:
                env_indexes = range(1 + vec_env_size, 2 * vec_env_size + 1)

            envs = [make_env(
                env_fn=env_fn, env_kwargs=env_kwargs,
                index_col_worker=index_col_worker,
                index_grad_worker=index_grad_worker,
                index_env=i, log_dir=log_dir, mode=mode,
                info_keywords=info_keywords
            ) for i in env_indexes]

            if len(envs) > 1:
                envs = SubprocVecEnv(envs)
            else:
                envs = DummyVecEnv(envs)

            envs = VecPyTorch(envs, device)

            envs.env_kwargs = env_kwargs

            return envs

        dummy_env = [make_env(
            env_fn=env_fn, env_kwargs=env_kwargs, index_col_worker=0,
            index_grad_worker=0, index_env=0)]
        dummy_env = DummyVecEnv(dummy_env)

        cls.action_space = dummy_env.action_space
        cls.observation_space = dummy_env.observation_space
        dummy_env.envs[0].close()

        return make_vec_env, dummy_env.action_space, dummy_env.observation_space



__version__ = "2.6.24"

from collections import namedtuple

__all__ = [
    "ALGORITHM", "EPISODES", "SCHEME", "TIME", "INFO_KEYS",
    "COLLECTION", "GRADIENT", "UPDATE", "OP_KEYS",
    "VERSION", "WEIGHTS", "NUMSAMPLES",
    "SYNC", "ASYNC", "CENTRAL", "PARALLEL", "FPS", "PL", "GA",
    "ACTOR", "STORAGE", "ENV_TRAIN", "ENV_TEST", "COMPONENT_KEYS",
    "OBS", "RHS", "DONE", "ACT", "REW", "OBS2", "RHS2", "DONE2",
    "DataTransitionKeys", "DataTransition", "OffPolicyDataKeys", "OnPolicyDataKeys",
]


# Information
ALGORITHM = "Algorithm"
EPISODES = "Episodes"
SCHEME = "Scheme"
TIME = "Time"
INFO_KEYS = (ALGORITHM, EPISODES, SCHEME, TIME)

# Operations
COLLECTION = "DataCollection"
GRADIENT = "GradientCompute"
UPDATE = "ActorUpdate"
PROCESSING = "DataProcessing"  # Includes Gradient computations + model Updates
CPRATIO = "DataCollection_DataProcessing_Ratio"
OP_KEYS = (COLLECTION, GRADIENT, UPDATE)

# Training
VERSION = "ActorVersion"
WEIGHTS = "ActorWeights"
NUMSAMPLES = "NumberSamples"

# Training Architecture
CENTRAL = "Central"
PARALLEL = "Parallel"
FPS = "FramesPerSecond"
PL = "PolicyLag"
GA = "GradientAsynchrony"
SYNC = "synchronous"
ASYNC = "asynchronous"

# Agent
ACTOR = "Actor"
STORAGE = "Storage"
ENV_TRAIN = "TrainEnvironment"
ENV_TEST = "TestEnvironment"
COMPONENT_KEYS = (ALGORITHM, ACTOR, STORAGE, ENV_TRAIN, ENV_TEST)

# DATA
OBS = "Observation"
RHS = "RecurrentHiddenStates"
DONE = "Done"
ACT = "Action"
ACTPROBS = "ActionProbs"
REW = "Reward"
IREW = "IntrinsicReward"
OBS2 = "NextObservation"
RHS2 = "NextRecurrentHiddenStates"
DONE2 = "NextDone"
VAL = "Value"
IVAL = "IntrinsicValue"
LOGP = "LogProbability"
ADV = "Advantage"
IADV = "IntrinsicAdvantage"
RET = "ExternalReturn"
IRET = "IntrinsicReturn"
MASK = "MaskedSamples"
EMBED = "Embedding"
INFO = "EnvironmentInformation"
DataTransitionKeys = (OBS, RHS, DONE, ACT, REW, OBS2, RHS2, DONE2, INFO)
DataTransition = namedtuple('DataTransition', DataTransitionKeys)
OffPolicyDataKeys = (OBS, RHS, DONE, ACT, REW, IREW, OBS2, RHS2, DONE2, ACTPROBS)
OnPolicyDataKeys = (OBS, RHS, DONE, ACT, REW, IREW, RET, IRET, VAL, IVAL, LOGP, ADV, IADV)
MBDataKeys = (OBS, RHS, DONE, ACT, REW, OBS2, RHS2, DONE2)
DemosDataKeys = (OBS, ACT, REW)

# DATA TYPES
float32 = "float32"
float16 = "float16"
int16 = "int16"
int8 = "int8"
uint8 = "uint8"

DEFAULT_DTYPES = {
    OBS: float32,
    RHS: float32,
    DONE: float32,
    ACT: float32,
    ACTPROBS: float32,
    REW: float32,
    OBS2: float32,
    RHS2: float32,
    DONE2: float32,
    VAL: float32,
    LOGP: float32,
    ADV: float32,
    RET: float32,
}

# ALGORITHMS
A2C = "A2C"
PPO = "PPO"
RND_PPO = "RND_PPO"
SAC = "SAC"
MPO = "MPO"
TD3 = "TD3"
DDPG = "DDPG"
DDQN = "DDQN"
MPC_RS = "ModelPredictiveControl_RandomShooting"




import gym
import pybulletgym
from pybulletgym.envs.roboschool.envs.env_bases import BaseBulletEnv
from pybulletgym.envs.roboschool.scenes.scene_bases import SingleRobotEmptyScene
from pybulletgym.envs.roboschool.robots.robot_bases import MJCFBasedRobot
import numpy as np


class SparseReacher(MJCFBasedRobot):
    TARG_LIMIT = 0.27

    def __init__(self):
        MJCFBasedRobot.__init__(self, 'reacher.xml', 'body0', action_dim=2, obs_dim=9)

    def robot_specific_reset(self, bullet_client):
        self.jdict["target_x"].reset_current_position(
            self.np_random.uniform(low=-self.TARG_LIMIT, high=self.TARG_LIMIT), 0)
        self.jdict["target_y"].reset_current_position(
            self.np_random.uniform(low=-self.TARG_LIMIT, high=self.TARG_LIMIT), 0)
        self.fingertip = self.parts["fingertip"]
        self.target = self.parts["target"]
        self.central_joint = self.jdict["joint0"]
        self.elbow_joint = self.jdict["joint1"]
        self.central_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)
        self.elbow_joint.reset_current_position(self.np_random.uniform(low=-3.14, high=3.14), 0)

    def apply_action(self, a):
        assert (np.isfinite(a).all())
        self.central_joint.set_motor_torque(0.05 * float(np.clip(a[0], -1, +1)))
        self.elbow_joint.set_motor_torque(0.05 * float(np.clip(a[1], -1, +1)))

    def calc_state(self):

        theta, self.theta_dot = self.central_joint.current_relative_position()
        self.gamma, self.gamma_dot = self.elbow_joint.current_relative_position()
        target_x, _ = self.jdict["target_x"].current_position()
        target_y, _ = self.jdict["target_y"].current_position()

        self.to_target_vec = np.array(self.fingertip.pose().xyz()) - np.array(self.target.pose().xyz())

        return np.array([
            target_x,
            target_y,
            self.to_target_vec[0],
            self.to_target_vec[1],
            np.cos(theta),
            np.sin(theta),
            self.theta_dot,
            self.gamma,
            self.gamma_dot,
        ])

    def calc_potential(self):
        return -100 * np.linalg.norm(self.to_target_vec)


class SparseReacherBulletEnv(BaseBulletEnv):
    def __init__(self):

        self.robot = SparseReacher()
        BaseBulletEnv.__init__(self, self.robot)

    def create_single_player_scene(self, bullet_client):
        return SingleRobotEmptyScene(bullet_client, gravity=0.0, timestep=0.0165, frame_skip=1)

    def step(self, a):
        assert (not self.scene.multiplayer)
        self.robot.apply_action(a)
        self.scene.global_step()

        state = self.robot.calc_state()

        potential_old = self.potential
        self.potential = self.robot.calc_potential()

        electricity_cost = (
                -0.10 * (np.abs(a[0] * self.robot.theta_dot) + np.abs(a[1] * self.robot.gamma_dot))  # work torque*angular_velocity
                - 0.01 * (np.abs(a[0]) + np.abs(a[1]))  # stall torque require some energy
        )

        stuck_joint_cost = -0.1 if np.abs(np.abs(self.robot.gamma) - 1) < 0.01 else 0.0
        self.rewards = [float(self.potential - potential_old), float(electricity_cost), float(stuck_joint_cost)]
        self.HUD(state, a, False)

        threshold = 1
        if self.potential > -threshold:
            reward = threshold + self.potential
        else:
            reward = 0

        return state, reward, False, {}

    def camera_adjust(self):
        x, y, z = self.robot.fingertip.pose().xyz()
        x *= 0.5
        y *= 0.5
        self.camera.move_and_look_at(0.3, 0.3, 0.3, x, y, z)


import gym
import numpy as np
import torch
import torch.nn as nn
from pytorchrl.agent.actors.utils import init
from transformers import OpenAIGPTConfig, OpenAIGPTModel


class GPT(nn.Module):
    """
    Wrapper to be able to use GPT model from transformers.

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    transformers_config : list
        Hidden layers sizes.
    """
    def __init__(self, input_space, transformers_config):
        super(GPT, self).__init__()

        # Define feature extractor
        self.feature_extractor = OpenAIGPTModel(transformers_config)

    def forward(self, inputs):
        """
        Forward pass Neural Network.

        This model assumes padding tokens are -1.

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out3 : torch.tensor
            Output feature map.
        """

        batch_size = inputs.size(0)
        inputs1 = inputs.view(batch_size, -1)
        inputs = inputs1 if len(inputs.shape) == 1 else inputs

        # Masks
        mask_attn = (inputs < 0.0)
        has_masked_tokens = (mask_attn == True).any()
        has_masked_tokens_in_all_sequences = mask_attn.any(dim=1).all() if batch_size > 1 else mask_attn.any()

        # Identify unnecessary padding
        if has_masked_tokens_in_all_sequences:
            ask_dim0, mask_dim1 = torch.where(mask_attn.all(dim=0, keepdims=True))
            last_idx = mask_dim1.min()
        else:
            last_idx = inputs.size(1) + 1

        # Remove -1 padding tokens
        inputs[mask_attn] = 0.0

        # Forward pass
        out1 = self.feature_extractor(
            input_ids=inputs1[:, 0:last_idx].long(),  # Shape (batch_size, sequence_length)
            attention_mask=1 - mask_attn[:, 0:last_idx].long(),  # Shape (batch_size, sequence_length)
        ).last_hidden_state

        # Prepare outputs
        if has_masked_tokens:  # Data collection
            idxs = (mask_attn.size(1) - mask_attn.sum(1)) - 1
            out2 = out1[torch.arange(len(out1)), idxs.view(-1)]
        else:  # Gradient computation
            out2 = out1[:, -1]

        # Place again padding tokens
        inputs[mask_attn] = - 1.0

        return out2



import pytorchrl as prl
from pytorchrl.scheme.gradients.g_worker import GWorker
from pytorchrl.scheme.base.worker_set import WorkerSet as WS
from pytorchrl.scheme.base.worker import default_remote_config


class GWorkerSet(WS):
    """
    Class to better handle the operations of ensembles of GWorkers.

    Parameters
    ----------
    num_workers : int
        Number of remote workers in the worker set.
    index_parent : int
        Worker index of parent gradient worker.
    local_device : str
        "cpu" or specific GPU "cuda:number`" to use for computation.
    col_execution : str
        Execution patterns for data collection.
    col_communication : str
        Communication coordination pattern for data collection.
    col_workers_factory : func
        A function that creates a set of data collection workers.
    compress_grads_to_send : bool
        Whether or not to compress gradients before sending then to the update worker.
    col_fraction_workers : float
        Minimum fraction of samples required to stop if collection is
        synchronously coordinated and most workers have finished their
        collection task.
    grad_worker_resources : dict
        Ray resource specs for the remote workers.
    initial_weights : ray object ID
        Initial model weights.

    Attributes
    ----------
    worker_class : python class
        Worker class to be instantiated to create Ray remote actors.
    remote_config : dict
        Ray resource specs for the remote workers.
    worker_params : dict
        Keyword arguments of the worker_class.
    num_workers : int
        Number of remote workers in the worker set.
    """

    def __init__(self,
                 num_workers,
                 index_parent,
                 local_device,
                 col_execution,
                 col_communication,
                 col_workers_factory,
                 col_fraction_workers,
                 grad_worker_resources,
                 compress_grads_to_send=False):

        self.worker_class = GWorker
        self.num_workers = num_workers
        default_remote_config.update(grad_worker_resources)
        self.remote_config = default_remote_config

        self.worker_params = {
            "col_execution": col_execution,
            "col_communication": col_communication,
            "col_workers_factory": col_workers_factory,
            "col_fraction_workers": col_fraction_workers,
            "compress_grads_to_send": compress_grads_to_send,
        }

        super(GWorkerSet, self).__init__(
            worker=self.worker_class,
            local_device=local_device,
            num_workers=self.num_workers,
            worker_params=self.worker_params,
            index_parent_worker=index_parent,
            worker_remote_config=self.remote_config)

    @classmethod
    def create_factory(cls,
                       num_workers,
                       col_workers_factory,
                       col_fraction_workers=1.0,
                       col_execution=prl.CENTRAL,
                       col_communication=prl.SYNC,
                       compress_grads_to_send=False,
                       grad_worker_resources=default_remote_config):
        """
        Returns a function to create new CWorkerSet instances.

        Parameters
        ----------
        num_workers : int
            Number of remote workers in the worker set.
        col_execution : str
            Execution patterns for data collection.
        col_communication : str
            Communication coordination pattern for data collection.
        col_workers_factory : func
            A function that creates a set of data collection workers.
        col_fraction_workers : float
            Minimum fraction of samples required to stop if collection is
            synchronously coordinated and most workers have finished their
            collection task.
        compress_grads_to_send : bool
            Whether or not to compress gradients before sending then to the update worker.
        grad_worker_resources : dict
            Ray resource specs for the remote workers.

        Returns
        -------
        grad_worker_set_factory : func
            creates a new GWorkerSet class instance.
        """
        def grad_worker_set_factory(device, index_parent):
            """
            Creates and returns a GWorkerSet class instance.

            Parameters
            ----------
            device : str
                "cpu" or specific GPU "cuda:number`" to use for computation.
            index_parent : int
                Worker index of parent gradient worker.

            Returns
            -------
            GWorkerSet : GWorkerSet
                A new GWorkerSet class instance.
            """
            return cls(
                local_device=device,
                num_workers=num_workers,
                index_parent=index_parent,
                col_execution=col_execution,
                col_communication=col_communication,
                col_workers_factory=col_workers_factory,
                col_fraction_workers=col_fraction_workers,
                grad_worker_resources=grad_worker_resources,
                compress_grads_to_send=compress_grads_to_send)

        return grad_worker_set_factory



import numpy as np
import torch
from collections import deque
import pytorchrl as prl
from pytorchrl.agent.storages.off_policy.per_buffer import PERBuffer as B


def dim0_reshape(tensor, size1, size2):
    """
    Reshapes tensor so indices are defined like this:

    00, 01, 02, 03, 04, 05, 06, 07, 08, 09, size + 1, ..., self.max_size
    10, 11, 12, 13, 14, 15, 16, 17, 18, 19, size + 1, ..., self.max_size
    20, 21, 22, 23, 24, 25, 26, 27, 28, 29, size + 1, ..., self.max_size

    """
    return np.moveaxis(tensor, [0, 1], [1, 0])[:, size1: size2].reshape(-1, *tensor.shape[2:])


class EREBuffer(B):
    """
    Storage class for Off-Policy algorithms with Emphasizing Recent Experience
    buffer (https://arxiv.org/abs/1906.04009).

    This component extends PERBuffer, allowing to combine ERE with Prioritized
    Experience Replay (PER) if required. Nonetheless PER parameters, epsilon,
    alpha and beta, are set by default to values that make PER equivalent to
    a vanilla replay buffer, allowing to use only ERE. Also n step learning
    can be combined with PER and ERE using this component, but default n_step
    value is 1.

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    n_step : int or float
        Number of future steps used to computed the truncated n-step return value.
    epsilon : float
        PER epsilon parameter.
    alpha : float
        PER alpha parameter.
    beta : float
        PER beta parameter.
    default_error : int or float
        Default TD error value to use for newly added data samples.
    eta : float
        ERE eta parameter.
    cmin : int
        ERE cmin parameter.
    """

    # Data fields to store in buffer and contained in the generated batches
    storage_tensors = prl.OffPolicyDataKeys

    def __init__(self, size, device, actor, algorithm, envs, n_step=1, epsilon=0.0, alpha=0.0, beta=1.0, default_error=1000000, eta=1.0, cmin=5000):

        super(EREBuffer, self).__init__(
            size=size, device=device, actor=actor, algorithm=algorithm,
            envs=envs, n_step=n_step, epsilon=epsilon, alpha=alpha, beta=beta,
            default_error=default_error)

        self.eta = eta
        self.cmin = cmin
        self.initial_eta = eta
        self.eps_reward = deque(maxlen=20)
        self.max_grad_rew, self.min_grad_rew = - np.Inf, 0.0

    @classmethod
    def create_factory(cls, size, n_step=1, epsilon=0.0, alpha=0.0, beta=1.0, default_error=1000000, eta=1.0, cmin=5000):
        """
        Returns a function that creates EREBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        n_step : int or float
            Number of future steps used to computed the truncated n-step return value.
        epsilon : float
            PER epsilon parameter.
        alpha : float
            PER alpha parameter.
        beta : float
            PER beta parameter.
        default_error : int or float
            Default TD error value to use for newly added data samples.
        eta : float
            ERE eta parameter.
        cmin : int
            ERE cmin parameter.

        Returns
        -------
        create_buffer_instance : func
            creates a new EREBuffer class instance.
        """

        def create_buffer(device, actor, algorithm, envs):
            """Create and return a EREBuffer instance."""
            return cls(size, device, actor, algorithm, envs, n_step, epsilon, alpha, beta, default_error, eta, cmin)

        return create_buffer

    def after_gradients(self, batch, info):
        """
        Steps required after updating actor policy model

        Parameters
        ----------
        batch : dict
            Data batch used to compute the gradients.
        info : dict
            Additional relevant info from gradient computation.

        Returns
        -------
        info : dict
            info dict updated with relevant info from Storage.
        """

        if "per_weights" in batch.keys() and isinstance(batch["per_weights"], torch.Tensor):

            num_proc = self.data[prl.DONE].shape[1]

            if self.recurrent_actor:

                N = self.size // self.sequence_length
                ck = batch.pop("ck")
                start = int((N - ck) * self.sequence_length)
                end = int(self.size + self.sequence_length)

                # Get data indices and td errors
                idxs = np.array(batch.pop("idxs")).reshape(-1, self.sequence_length)
                errors = info[prl.ALGORITHM]["errors"].reshape(-1, self.non_overlap_length)

                # Since sequences overlap, update both current sequence and
                # start of the next overlapping sequence
                idxs = np.concatenate([idxs, idxs[:, :self.overlap_length] + self.overlap_length], axis=1)
                errors = torch.cat([errors, errors[:, :self.overlap_length]], dim=1)

                for i, e in zip(idxs, errors):

                    # Assign priorities to both end of current sequence
                    # and start of next sequence
                    dim0_reshape(self.data["priority"], start, end
                                 )[i] = self.get_priority(e.unsqueeze(1))

                    # Update current sequence average priority
                    sequence = dim0_reshape(self.data["priority"], start, end)[
                        i - self.overlap_length]
                    pri = self.get_sequence_priority(sequence)
                    dim0_reshape(self.data["priority"], start, end)[i -
                        self.overlap_length] = pri * np.ones(sequence.shape)

                    # Update next sequence average if some overlap
                    if self.overlap_length > 0:
                        sequence = dim0_reshape(self.data["priority"],
                            start, end)[i + self.non_overlap_length]
                        pri = self.get_sequence_priority(sequence)
                        dim0_reshape(self.data["priority"], start, end)[i +
                            self.non_overlap_length] = pri * np.ones(sequence.shape)

            else:
                N = self.size * num_proc
                ck = batch.pop("ck")
                self.data["priority"][0:self.size].reshape(-1, *self.data[
                    "priority"].shape[2:])[N - ck: N][batch.pop("idxs")] = \
                        self.get_priority(info[prl.ALGORITHM]["errors"])

        info = self.update_eta(info)
        return info

    def update_eta(self, info):
        """
        Adjust eta parameter based on how fast or slow the agent is learning
        in recent episodes
        """

        slopes = np.linspace(-1.0, 1.0, 1000)
        etas = np.linspace(1.0, self.initial_eta, 1000)
        if prl.EPISODES in info.keys() and "TrainReward" in info[prl.EPISODES].keys():
            self.eps_reward.append(info[prl.EPISODES]["TrainReward"])

        if len(self.eps_reward) == self.eps_reward.maxlen:
            reward_grad = np.gradient(self.eps_reward).mean()

            if reward_grad > self.max_grad_rew: self.max_grad_rew = reward_grad
            else: self.max_grad_rew *= 0.9999

            if reward_grad < - self.min_grad_rew: self.min_grad_rew = abs(reward_grad)
            else: self.min_grad_rew *= 0.9999

            if np.sign(reward_grad) == 1: reward_grad /= self.max_grad_rew
            else: reward_grad /= self.min_grad_rew

            idx = (np.abs(slopes - reward_grad)).argmin()
            self.eta = etas[idx]
            info[prl.ALGORITHM].update({"RewardGradient": reward_grad})
            info[prl.ALGORITHM].update({"eta": self.eta})

        return info

    def generate_batches(self, num_mini_batch, mini_batch_size, num_epochs=1):
        """
        Returns a batch iterator to update actor.

        Parameters
        ----------
        num_mini_batch : int
           Number mini batches per epoch.
        mini_batch_size : int
            Number of samples contained in each mini batch.
        num_epochs : int
            Number of epochs.

        Yields
        ------
        batch : dict
            Generated data batches. Contains also extra information relevant to ERE.
        """
        num_proc = self.data[prl.DONE].shape[1]

        if self.recurrent_actor:  # Batches to a feed recurrent actor

            for k in range(num_mini_batch):

                # Define batch structure
                batch = {k: {} if not isinstance(self.data[k], dict) else
                {x: {} for x in self.data[k]} for k in self.data.keys()}

                sequences_x_batch = mini_batch_size // self.sequence_length + 1
                sequences_x_proc = int(self.size / self.sequence_length)
                N = sequences_x_proc

                assert self.size % self.sequence_length == 0, \
                    "Buffer does not contain an integer number of complete rollout sequences"

                if num_proc * self.size < self.cmin:  # Standard

                    seq_idxs = np.random.randint(0, num_proc * sequences_x_proc, size=sequences_x_batch)
                    per_weigths, ck = 1.0, N
                    start, end = int((N - ck) * self.sequence_length), int(self.size)

                elif self.alpha == 0.0:  # ERE

                    cmin = self.cmin // self.sequence_length // num_proc
                    ck = int(max(N * self.eta ** ((1000 * k) / num_mini_batch), cmin))

                    # reshape between N - ck, N
                    start, end = int((N - ck) * self.sequence_length), int(self.size)

                    seq_idxs = np.random.randint(0, ck * num_proc, size=sequences_x_batch)
                    per_weigths = 1.0

                else:  # PER + ERE

                    cmin = self.cmin // self.sequence_length // num_proc
                    ck = max(N * self.eta ** ((1000 * k) / num_mini_batch), cmin)

                    # reshape between N - ck, N
                    start = int((N - ck) * self.sequence_length)
                    end = int(self.size)

                    priors = dim0_reshape(self.data["priority"], start, end)
                    probs = priors / priors.sum()

                    per_weigths = np.power(num_proc * self.size * probs, - self.beta)
                    per_weigths = per_weigths / per_weigths.max()

                    # Trick to allow updating priorities of next overlapping
                    # sequences after gradient computation. Insert some "0.0"
                    # values in per_weights for end-of-row + 1 sequences.
                    per_weigths = np.split(per_weigths, num_proc)
                    per_weigths = [np.concatenate([chunk, np.zeros((
                        self.sequence_length, 1))]) for chunk in per_weigths]
                    per_weigths = np.concatenate(per_weigths)

                    # Trick to allow updating priorities of next overlapping
                    # sequences after gradient computation. Insert some "0.0"
                    # values in probs for end-of-row + 1 sequences.
                    probs = probs[int((N - ck + 1) * self.sequence_length) - 1
                        ::self.sequence_length] * self.sequence_length
                    probs = np.split(probs, num_proc)
                    probs = [np.concatenate([chunk, np.zeros((1, 1))]) for chunk in probs]
                    ext_probs = np.concatenate(probs).squeeze(1)

                    end += self.sequence_length
                    seq_idxs = np.random.choice(range(
                        len(ext_probs)), size=sequences_x_batch, p=ext_probs)

                # Get data indexes
                idxs = []
                for idx in seq_idxs:
                    idxs += range(idx * self.sequence_length, (idx + 1) * self.sequence_length)

                if not isinstance(per_weigths, float):
                    per_weigths = torch.as_tensor(per_weigths[idxs], dtype=torch.float32).to(self.device)

                # Fill up batch with data
                for k, v in self.data.items():
                    # Only first recurrent state in each sequence needed
                    positions = seq_idxs * self.sequence_length if k in (prl.RHS, prl.RHS2) else idxs
                    if isinstance(v, dict):
                        for x, y in v.items():
                            t = dim0_reshape(y, start, end)[positions]
                            batch[k][x] = torch.as_tensor(t, dtype=torch.float32).to(self.device)
                    else:
                        t = dim0_reshape(v, start, end)[positions]
                        batch[k] = torch.as_tensor(t, dtype=torch.float32).to(self.device)

                batch.update({"per_weights": per_weigths, "n_step": self.n_step, "idxs": idxs, "ck": ck})
                yield batch

        else:  # Batches for a feed forward actor
            for k in range(num_mini_batch):

                batch = {k: {} for k in self.data.keys()}
                N = num_proc * self.size
                per_weigths = None

                if num_proc * self.size < self.cmin:  # Standard
                    samples = np.random.randint(0, num_proc * self.size, size=mini_batch_size)
                    ck = N

                elif self.alpha == 0.0:  # ERE
                    ck = int(max(N * self.eta ** ((1000 * k) / num_mini_batch), self.cmin))
                    samples = np.random.randint(ck, size=mini_batch_size)

                else:  # PER + ERE
                    ck = int(max(N * self.eta ** ((1000 * k) / num_mini_batch), self.cmin))
                    priors = self.data["priority"][0:self.size].reshape(-1, 1)[N - ck: N]
                    probs = priors / priors.sum()
                    samples = np.random.choice(range(ck), size=mini_batch_size, p=probs.squeeze(1))

                    per_weigths = np.power(ck * probs, - self.beta)
                    per_weigths /= per_weigths.max()
                    per_weigths = per_weigths[samples]
                    per_weigths = torch.as_tensor(per_weigths, dtype=torch.float32).to(self.device)

                for k, v in self.data.items():

                    if k in (prl.RHS, prl.RHS2):
                        size, idxs = 1, np.array([0])
                    else:
                        size, idxs = self.size, samples

                    if isinstance(v, dict):
                        for x, y in v.items():
                            batch[k][x] = torch.as_tensor(y[0:self.size].reshape(
                                -1, *y.shape[2:])[N - ck: N][idxs], dtype=torch.float32).to(self.device)
                    else:
                        batch[k] = torch.as_tensor(v[0:self.size].reshape(
                            -1, *v.shape[2:])[N - ck: N][idxs], dtype=torch.float32).to(self.device)

                batch.update({"n_step": self.n_step, "idxs": idxs, "ck": ck})

                if per_weigths is not None:
                    batch.update({"per_weights": per_weigths})

                yield batch





"""
Helpers for dealing with vectorized environments.
"""

from collections import OrderedDict

import gym
import numpy as np


def tile_images(img_nhwc):
    """
    Tile N images into one big PxQ image
    (P,Q) are chosen to be as close as possible, and if N
    is square, then P=Q.
    input: img_nhwc, list or array of images, ndim=4 once turned into array
        n = batch index, h = height, w = width, c = channel
    returns:
        bigim_HWc, ndarray with ndim=3
    """
    img_nhwc = np.asarray(img_nhwc)
    N, h, w, c = img_nhwc.shape
    H = int(np.ceil(np.sqrt(N)))
    W = int(np.ceil(float(N)/H))
    img_nhwc = np.array(list(img_nhwc) + [img_nhwc[0]*0 for _ in range(N, H*W)])
    img_HWhwc = img_nhwc.reshape(H, W, h, w, c)
    img_HhWwc = img_HWhwc.transpose(0, 2, 1, 3, 4)
    img_Hh_Ww_c = img_HhWwc.reshape(H*h, W*w, c)
    return img_Hh_Ww_c


def copy_obs_dict(obs):
    """
    Deep-copy an observation dict.
    """
    return {k: np.copy(v) for k, v in obs.items()}


def dict_to_obs(obs_dict):
    """
    Convert an observation dict into a raw array if the
    original observation space was not a Dict space.
    """
    if set(obs_dict.keys()) == {None}:
        return obs_dict[None]
    return obs_dict


def obs_space_info(obs_space):
    """
    Get dict-structured information about a gym.Space.
    Returns:
      A tuple (keys, shapes, dtypes):
        keys: a list of dict keys.
        shapes: a dict mapping keys to shapes.
        dtypes: a dict mapping keys to dtypes.
    """
    if isinstance(obs_space, gym.spaces.Dict):
        assert isinstance(obs_space.spaces, OrderedDict)
        subspaces = obs_space.spaces
    elif isinstance(obs_space, gym.spaces.Tuple):
        assert isinstance(obs_space.spaces, tuple)
        subspaces = {i: obs_space.spaces[i] for i in range(len(obs_space.spaces))}
    else:
        subspaces = {None: obs_space}
    keys = []
    shapes = {}
    dtypes = {}
    for key, box in subspaces.items():
        keys.append(key)
        shapes[key] = box._shape if hasattr(box, "_shape") else box.shape
        dtypes[key] = box.dtype
    return keys, shapes, dtypes


def obs_to_dict(obs):
    """
    Convert an observation into a dict.
    """
    if isinstance(obs, dict):
        return obs
    return {None: obs}



from abc import ABC, abstractmethod
import torch


class PolicyLossAddOn(ABC):
    """Base class for all add ons to the policy loss."""

    @abstractmethod
    def setup(self, actor, device):
        """Initializes the class."""
        raise NotImplementedError

    @abstractmethod
    def compute_loss_term(self, batch, dist_entropy=None):
        """Calculates addon loss term."""
        raise NotImplementedError



import torch
import torch.nn as nn


class GruNet(nn.Module):
    """Implements a GRU model"""
    def __init__(self, input_size, output_size=256, num_layers=1, activation=nn.ReLU, dropout=0.0):
        """
        Initializes a GRU model.

        Parameters
        ----------
        input_size : int
            Input feature map size.
        output_size : int
            Recurrent hidden state and output size.
        dropout : float
            If non-zero, introduces a Dropout layer on the outputs of each GRU layer except the last layer.
        num_layers : int
            Number of recurrent layers.
        activation : func
            Non-linear activation function.
        """
        super(GruNet, self).__init__()

        self._rnn = nn.GRU(input_size=input_size, num_layers=num_layers, hidden_size=output_size, dropout=dropout)
        self._num_outputs = output_size
        self.activation = activation()

        self.train()

    @property
    def num_outputs(self):
        """Output feature map size (as in np.prod(self.output_shape))."""
        return self._num_outputs

    @property
    def recurrent_hidden_state_size(self):
        """Recurrent hidden state size"""
        return self._num_outputs

    def _forward_gru(self, x, hxs, done):
        """
        Fast forward pass GRU network.
        from Ilya Kostrikov. PyTorch Implementations of Reinforcement Learning Algorithms.
        https://github.com/ikostrikov/pytorch-a2c-ppo-acktr-gail. 2018

        Parameters
        ----------
        x : torch.tensor
            Feature map obtained from environment observation.
        hxs : torch.tensor
            Current recurrent hidden state.
        done : torch.tensor
            Current done tensor, indicating if episode has finished.

        Returns
        -------
        x : torch.tensor
            Feature map obtained after GRU.
        hxs : torch.tensor
            Updated recurrent hidden state.
        """

        masks = 1 - done

        if x.size(0) == hxs.size(0):
            self._rnn.flatten_parameters()
            x, hxs = self._rnn(x.unsqueeze(0), (hxs * masks).unsqueeze(0))
            x = x.squeeze(0)
            hxs = hxs.squeeze(0)
        else:
            # x is a (T, N, -1) tensor that has been flatten to (T * N, -1)
            N = hxs.size(0)
            T = int(x.size(0) / N)

            # unflatten
            x = x.view(T, N, x.size(1))

            # Same deal with masks
            masks = masks.view(T, N)

            # Let's figure out which steps in the sequence have a zero for any agent
            # We will always assume t=0 has a zero in it as that makes the logic cleaner
            has_zeros = torch.nonzero(((masks[1:] == 0.0).any(dim=-1)), as_tuple=False).squeeze().cpu()

            # +1 to correct the masks[1:]
            if has_zeros.dim() == 0:
                # Deal with scalar
                has_zeros = [has_zeros.item() + 1]
            else:
                has_zeros = (has_zeros + 1).numpy().tolist()

            # add t=0 and t=T to the list
            has_zeros = [0] + has_zeros + [T]

            hxs = hxs.unsqueeze(0)
            outputs = []
            for i in range(len(has_zeros) - 1):
                # We can now process steps that don't have any zeros in masks together!
                # This is much faster
                start_idx = has_zeros[i]
                end_idx = has_zeros[i + 1]

                self._rnn.flatten_parameters()
                rnn_scores, hxs = self._rnn(
                    x[start_idx:end_idx],
                    hxs * masks[start_idx].view(1, -1, 1))

                outputs.append(rnn_scores)

            # x is a (T, N, -1) tensor
            x = torch.cat(outputs, dim=0)
            # flatten
            x = x.view(T * N, -1)
            hxs = hxs.squeeze(0)

        return x, hxs

    def forward(self, inputs, rhs, done):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            A tensor containing episode observations.
        rhs : torch.tensor
            A tensor representing the recurrent hidden states.
        done : torch.tensor
            A tensor indicating where episodes end.

        Returns
        -------
        x : torch.tensor
            Output feature map.
        rhs : torch.tensor
            Updated recurrent hidden state.
        """

        x = inputs.view(inputs.size(0), -1)
        x, rhs = self._forward_gru(x, rhs, done)
        # x = self.activation(x)
        assert len(x.shape) == 2 and x.shape[1] == self.num_outputs
        return x, rhs

    def get_initial_recurrent_state(self, num_proc):
        """Returns a tensor of zeros with the expected shape of the model's rhs."""
        return torch.zeros(num_proc, self._layer_size)




import itertools
import numpy as np
from copy import deepcopy

import torch
import torch.nn as nn
import torch.optim as optim

import pytorchrl as prl
from pytorchrl.agent.algorithms.base import Algorithm
from pytorchrl.agent.algorithms.utils import get_gradients, set_gradients
from pytorchrl.agent.algorithms.policy_loss_addons import PolicyLossAddOn


class SAC(Algorithm):
    """
    Soft Actor Critic algorithm class.

    Algorithm class to execute SAC, from Haarnoja et al.
    (https://arxiv.org/abs/1812.05905). Algorithms are modules generally
    required by multiple workers, so SAC.algo_factory(...) returns a function
    that can be passed on to workers to instantiate their own SAC module.

    Parameters
    ----------
    device : torch.device
        CPU or specific GPU where class computations will take place.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor_critic class instance.
    lr_pi : float
        Policy optimizer learning rate.
    lr_q : float
        Q-nets optimizer learning rate.
    lr_alpha : float
        Alpha optimizer learning rate.
    gamma : float
        Discount factor parameter.
    initial_alpha : float
        Initial entropy coefficient value (temperature).
    polyak : float
        SAC polyak averaging parameter.
    num_updates : int
        Num consecutive actor updates before data collection continues.
    update_every : int
        Regularity of actor updates in number environment steps.
    start_steps : int
        Num of initial random environment steps before learning starts.
    mini_batch_size : int
        Size of actor update batches.
    target_update_interval : float
        regularity of target nets updates with respect to actor Adam updates.
    num_test_episodes : int
        Number of episodes to complete in each test phase.
    test_every : int
        Regularity of test evaluations in actor updates.
    max_grad_norm : float
        Gradient clipping parameter.
    policy_loss_addons : list
        List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

    Examples
    --------
    >>> create_algo = SAC.create_factory(
            lr_q=1e-4, lr_pi=1e-4, lr_alpha=1e-4, gamma=0.99, polyak=0.995,
            num_updates=50, update_every=50, test_every=5000, start_steps=20000,
            mini_batch_size=64, alpha=1.0, num_test_episodes=0, target_update_interval=1)
    """

    def __init__(self,
                 device,
                 envs,
                 actor,
                 lr_q=1e-4,
                 lr_pi=1e-4,
                 lr_alpha=1e-4,
                 gamma=0.99,
                 polyak=0.995,
                 num_updates=1,
                 update_every=50,
                 test_every=1000,
                 max_grad_norm=0.5,
                 initial_alpha=1.0,
                 start_steps=20000,
                 mini_batch_size=64,
                 num_test_episodes=5,
                 target_update_interval=1,
                 policy_loss_addons=[]):

        # ---- General algo attributes ----------------------------------------

        # Discount factor
        self._gamma = gamma

        # Number of steps collected with initial random policy
        self._start_steps = int(start_steps)

        # Times data in the buffer is re-used before data collection proceeds
        self._num_epochs = int(1) # Default to 1 for off-policy algorithms

        # Number of data samples collected between network update stages
        self._update_every = int(update_every)

        # Number mini batches per epoch
        self._num_mini_batch = int(num_updates)

        # Size of update mini batches
        self._mini_batch_size = int(mini_batch_size)

        # Number of network updates between test evaluations
        self._test_every = int(test_every)

        # Number of episodes to complete when testing
        self._num_test_episodes = int(num_test_episodes)

        # ---- SAC-specific attributes ----------------------------------------

        self.iter = 0
        self.envs = envs
        self.actor = actor
        self.polyak = polyak
        self.device = device
        self.max_grad_norm = max_grad_norm
        self.target_update_interval = target_update_interval

        assert hasattr(self.actor, "q1"), "SAC requires double q critic (num_critics=2)"
        assert hasattr(self.actor, "q2"), "SAC requires double q critic (num_critics=2)"

        self.log_alpha = torch.tensor(
            data=[np.log(initial_alpha)], dtype=torch.float32,
            requires_grad=True, device=device)
        self.alpha = self.log_alpha.detach().exp()

        # Compute target entropy
        target_entropy = self.calculate_target_entropy()
        self.target_entropy = torch.tensor(
            data=target_entropy, dtype=torch.float32,
            requires_grad=False, device=device)

        # Create target networks
        self.actor_targ = deepcopy(actor)

        # Freeze target networks with respect to optimizers
        for p in self.actor_targ.parameters():
            p.requires_grad = False

        # List of parameters for both Q-networks
        q_params = itertools.chain(self.actor.q1.parameters(), self.actor.q2.parameters())

        # List of parameters for policy network
        p_params = itertools.chain(self.actor.policy_net.parameters())

        # ----- Policy Loss Addons --------------------------------------------

        # Sanity check, policy_loss_addons is a PolicyLossAddOn instance
        # or a list of PolicyLossAddOn instances
        assert isinstance(policy_loss_addons, (PolicyLossAddOn, list)),\
            "SAC policy_loss_addons parameter should be a  PolicyLossAddOn instance " \
            "or a list of PolicyLossAddOn instances"
        if isinstance(policy_loss_addons, list):
            for addon in policy_loss_addons:
                assert isinstance(addon, PolicyLossAddOn), \
                    "SAC policy_loss_addons parameter should be a  PolicyLossAddOn " \
                    "instance or a list of PolicyLossAddOn instances"
        else:
            policy_loss_addons = [policy_loss_addons]

        self.policy_loss_addons = policy_loss_addons
        for addon in self.policy_loss_addons:
            addon.setup(self.actor, self.device)

        # ----- Optimizers ----------------------------------------------------

        self.pi_optimizer = optim.Adam(p_params, lr=lr_pi)
        self.q_optimizer = optim.Adam(q_params, lr=lr_q)
        self.alpha_optimizer = optim.Adam([self.log_alpha], lr=lr_alpha)

    @classmethod
    def create_factory(cls,
                       lr_q=1e-4,
                       lr_pi=1e-4,
                       lr_alpha=1e-4,
                       gamma=0.99,
                       polyak=0.995,
                       num_updates=50,
                       test_every=5000,
                       update_every=50,
                       start_steps=1000,
                       max_grad_norm=0.5,
                       initial_alpha=1.0,
                       mini_batch_size=64,
                       num_test_episodes=5,
                       target_update_interval=1.0,
                       policy_loss_addons=[]):
        """
        Returns a function to create new SAC instances.

        Parameters
        ----------
        lr_pi : float
            Policy optimizer learning rate.
        lr_q : float
            Q-nets optimizer learning rate.
        lr_alpha : float
            Alpha optimizer learning rate.
        gamma : float
            Discount factor parameter.
        initial_alpha : float
            Initial entropy coefficient value.
        polyak : float
            SAC polyak averaging parameter.
        num_updates : int
            Num consecutive actor updates before data collection continues.
        update_every : int
            Regularity of actor updates in number environment steps.
        start_steps : int
            Num of initial random environment steps before learning starts.
        mini_batch_size : int
            Size of actor update batches.
        target_update_interval : float
            regularity of target nets updates with respect to actor Adam updates.
        num_test_episodes : int
            Number of episodes to complete in each test phase.
        test_every : int
            Regularity of test evaluations in actor updates.
        max_grad_norm : float
            Gradient clipping parameter.
        policy_loss_addons : list
            List of PolicyLossAddOn components adding loss terms to the algorithm policy loss.

        Returns
        -------
        create_algo_instance : func
            Function that creates a new SAC class instance.
        algo_name : str
            Name of the algorithm.
        """

        def create_algo_instance(device, actor, envs):
            return cls(lr_q=lr_q,
                       lr_pi=lr_pi,
                       lr_alpha=lr_alpha,
                       envs=envs,
                       actor=actor,
                       gamma=gamma,
                       device=device,
                       polyak=polyak,
                       test_every=test_every,
                       start_steps=start_steps,
                       num_updates=num_updates,
                       update_every=update_every,
                       initial_alpha=initial_alpha,
                       max_grad_norm=max_grad_norm,
                       mini_batch_size=mini_batch_size,
                       num_test_episodes=num_test_episodes,
                       target_update_interval=target_update_interval,
                       policy_loss_addons=policy_loss_addons)

        return create_algo_instance, prl.SAC

    @property
    def gamma(self):
        """Returns discount factor gamma."""
        return self._gamma

    @property
    def start_steps(self):
        """Returns the number of steps to collect with initial random policy."""
        return self._start_steps

    @property
    def num_epochs(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_epochs

    @property
    def update_every(self):
        """
        Returns the number of data samples collected between
        network update stages.
        """
        return self._update_every

    @property
    def num_mini_batch(self):
        """
        Returns the number of times the whole buffer is re-used before data
        collection proceeds.
        """
        return self._num_mini_batch

    @property
    def mini_batch_size(self):
        """
        Returns the number of mini batches per epoch.
        """
        return self._mini_batch_size

    @property
    def test_every(self):
        """Number of network updates between test evaluations."""
        return self._test_every

    @property
    def num_test_episodes(self):
        """
        Returns the number of episodes to complete when testing.
        """
        return self._num_test_episodes

    @property
    def discrete_version(self):
        """Returns True if action_space is discrete."""
        return self.actor.action_space.__class__.__name__ == "Discrete"

    def acting_step(self, obs, rhs, done, deterministic=False):
        """
        SAC acting function.

        Parameters
        ----------
        obs : torch.tensor
            Current world observation
        rhs : torch.tensor
            RNN recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        done : torch.tensor
            1.0 if current obs is the last one in the episode, else 0.0.
        deterministic : bool
            Whether to randomly sample action from predicted distribution or taking the mode.

        Returns
        -------
        action : torch.tensor
            Predicted next action.
        clipped_action : torch.tensor
            Predicted next action (clipped to be within action space).
        rhs : torch.tensor
            Policy recurrent hidden state (if policy is not a RNN, rhs will contain zeroes).
        other : dict
            Additional SAC predictions, which are not used in other algorithms.
        """

        with torch.no_grad():
            (action, clipped_action, logp_action, rhs,
             entropy_dist, dist) = self.actor.get_action(
                obs, rhs, done, deterministic=deterministic)

        return action, clipped_action, rhs, {}

    def compute_loss_q(self, batch, n_step=1, per_weights=1):
        """
        Calculate SAC Q-nets loss

        Parameters
        ----------
        batch : dict
            Data batch dict containing all required tensors to compute SAC losses.
        n_step : int or float
            Number of future steps used to computed the truncated n-step return value.
        per_weights :
            Prioritized Experience Replay (PER) important sampling weights or 1.0.

        Returns
        -------
        loss_q1 : torch.tensor
            Q1-net loss.
        loss_q2 : torch.tensor
            Q2-net loss.
        loss_q : torch.tensor
            Weighted average of loss_q1 and loss_q2.
        errors : torch.tensor
            TD errors.
        """

        o, rhs, d = batch[prl.OBS], batch[prl.RHS], batch[prl.DONE]
        a, r = batch[prl.ACT], batch[prl.REW]
        o2, rhs2, d2 = batch[prl.OBS2], batch[prl.RHS2], batch[prl.DONE2]

        if self.discrete_version:

            # Q-values for all actions
            q_scores = self.actor.get_q_scores(o, rhs, d)
            q1 = q_scores.get("q1").gather(1, a.long())
            q2 = q_scores.get("q2").gather(1, a.long())

            # Bellman backup for Q functions
            with torch.no_grad():

                # Target actions come from *current* policy
                a2, _, _, _, _, dist = self.actor.get_action(o2, rhs2, d2)

                bs, n = o.shape[0], dist.probs.shape[-1]
                actions = torch.arange(n)[..., None].expand(-1, bs).to(self.device)
                p_a2 = dist.expand((n, bs)).log_prob(actions).exp().transpose(0, 1)
                # p_a2 = dist.probs

                z = (p_a2 == 0.0).float() * 1e-8
                logp_a2 = torch.log(p_a2 + z)

                # Target Q-values
                q_scores_targ = self.actor_targ.get_q_scores(o2, rhs2, d2)
                q1_targ = q_scores_targ.get("q1")
                q2_targ = q_scores_targ.get("q2")

                q_targ = (p_a2 * (torch.min(q1_targ, q2_targ) - self.alpha * logp_a2)).sum(dim=1, keepdim=True)

                assert r.shape == q_targ.shape
                backup = r + (self.gamma ** n_step) * (1 - d2) * q_targ

        else:

            # Q-values for all actions
            q_scores = self.actor.get_q_scores(o, rhs, d, a)
            q1 = q_scores.get("q1")
            q2 = q_scores.get("q2")

            # Bellman backup for Q functions
            with torch.no_grad():

                # Target actions come from *current* policy
                a2, _, logp_a2, _, _, dist = self.actor.get_action(o2, rhs2, d2)

                # Target Q-values
                q_scores_targ = self.actor_targ.get_q_scores(o2, rhs2, d2, a2)
                q1_targ = q_scores_targ.get("q1")
                q2_targ = q_scores_targ.get("q2")

                q_pi_targ = torch.min(q1_targ, q2_targ)
                backup = r + (self.gamma ** n_step) * (1 - d2) * (q_pi_targ - self.alpha * logp_a2)

        # MSE loss against Bellman backup
        loss_q1 = (((q1 - backup) ** 2) * per_weights).mean()
        loss_q2 = (((q2 - backup) ** 2) * per_weights).mean()
        loss_q = 0.5 * loss_q1 + 0.5 * loss_q2

        # errors = (torch.min(q1, q2) - backup).abs().detach().cpu()
        # errors = torch.max((q1 - backup).abs(), (q2 - backup).abs()).detach().cpu()
        errors = (0.5 * (q1 - backup).abs() + 0.5 * (q2 - backup).abs()).detach().cpu()

        return loss_q1, loss_q2, loss_q, errors

    def compute_loss_pi(self, batch, per_weights=1):
        """
        Calculate SAC policy loss.

        Parameters
        ----------
        batch : dict
            Data batch dict containing all required tensors to compute SAC losses.
        per_weights :
            Prioritized Experience Replay (PER) important sampling weights or 1.0.

        Returns
        -------
        loss_pi : torch.tensor
            SAC policy loss.
        logp_pi : torch.tensor
            Log probability of predicted next action.
        """

        o, rhs, d = batch[prl.OBS], batch[prl.RHS], batch[prl.DONE]

        if self.discrete_version:

            pi, _, _, _, _, dist = self.actor.get_action(o, rhs, d)

            # Get action log probs
            bs, n = o.shape[0], dist.probs.shape[-1]
            actions = torch.arange(n)[..., None].expand(-1, bs).to(self.device)
            p_pi = dist.expand((n, bs)).log_prob(actions).exp().transpose(0, 1)
            # p_pi = dist.probs

            z = (p_pi == 0.0).float() * 1e-8
            logp_pi = torch.log(p_pi + z)
            logp_pi = torch.sum(p_pi * logp_pi, dim=1, keepdim=True)

            q_scores = self.actor.get_q_scores(o, rhs, d)
            q1 = q_scores.get("q1")
            q2 = q_scores.get("q2")
            q_pi = torch.sum(torch.min(q1, q2) * p_pi, dim=1, keepdim=True)

        else:

            pi, _, logp_pi, _, _, dist = self.actor.get_action(o, rhs, d)
            q_scores = self.actor.get_q_scores(o, rhs, d, pi)
            q1 = q_scores.get("q1")
            q2 = q_scores.get("q2")
            q_pi = torch.min(q1, q2)

        loss_pi = ((self.alpha * logp_pi - q_pi) * per_weights).mean()

        # Extend policy loss with addons
        addons_info = {}
        for addon in self.policy_loss_addons:
            addon_loss, addons_info = addon.compute_loss_term(data, dist, addons_info)
            loss_pi += addon_loss

        return loss_pi, logp_pi, addons_info

    def compute_loss_alpha(self, log_probs, per_weights=1):
        """
        Calculate SAC entropy loss.

        Parameters
        ----------
        log_probs : torch.tensor
            Log probability of predicted next action.
        per_weights :
            Prioritized Experience Replay (PER) important sampling weights or 1.0.

        Returns
        -------
        alpha_loss : torch.tensor
            SAC entropy loss.
        """
        alpha_loss = - ((self.log_alpha * (log_probs + self.target_entropy).detach()) * per_weights).mean()
        return alpha_loss

    def calculate_target_entropy(self):
        """Calculate SAC target entropy"""
        if self.discrete_version:
            target = - np.log(1.0 / self.actor.action_space.n) * 0.98
        else:
            target_old = - self.actor.action_space.shape[0]
            target = - np.prod(self.actor.action_space.shape)
            assert target_old == target
        return target

    def compute_gradients(self, batch, grads_to_cpu=True):
        """
        Compute loss and compute gradients but don't do optimization step,
        return gradients instead.

        Parameters
        ----------
        batch : dict
            data batch containing all required tensors to compute SAC losses.
        grads_to_cpu : bool
            If gradient tensor will be sent to another node, need to be in CPU.

        Returns
        -------
        grads : list of tensors
            List of actor gradients.
        info : dict
            Dict containing current SAC iteration information.
        """

        # Recurrent burn-in
        if self.actor.is_recurrent:
            batch = self.actor.burn_in_recurrent_states(batch)

        # N-step returns
        n_step = batch["n_step"] if "n_step" in batch else 1.0

        # PER
        per_weights = batch["per_weights"] if "per_weights" in batch else 1.0

        # Run one gradient descent step for Q1 and Q2
        loss_q1, loss_q2, loss_q, errors = self.compute_loss_q(batch, n_step, per_weights)
        self.q_optimizer.zero_grad()
        loss_q.backward(retain_graph=True)
        nn.utils.clip_grad_norm_(itertools.chain(
            self.actor.q1.parameters(), self.actor.q2.parameters()), self.max_grad_norm)
        q_grads = get_gradients(self.actor.q1, self.actor.q2, grads_to_cpu=grads_to_cpu)

        # Freeze Q-networks so you don't waste computational effort
        # computing gradients for them during the policy learning step.
        for p in itertools.chain(self.actor.q1.parameters(), self.actor.q2.parameters()):
            p.requires_grad = False

        # Run one gradient descent step for pi.
        loss_pi, logp_pi, addons_info = self.compute_loss_pi(batch, per_weights)
        self.pi_optimizer.zero_grad()
        loss_pi.backward()
        nn.utils.clip_grad_norm_(self.actor.policy_net.parameters(), self.max_grad_norm)
        pi_grads = get_gradients(self.actor.policy_net, grads_to_cpu=grads_to_cpu)

        for p in itertools.chain(self.actor.q1.parameters(), self.actor.q2.parameters()):
            p.requires_grad = True

        # Run one gradient descent step for alpha.
        self.alpha_optimizer.zero_grad()
        loss_alpha = self.compute_loss_alpha(logp_pi, per_weights)
        loss_alpha.backward()

        info = {
            "loss_q1": loss_q1.detach().item(),
            "loss_q2": loss_q2.detach().item(),
            "loss_pi": loss_pi.detach().item(),
            "loss_alpha": loss_alpha.detach().item(),
            "alpha": self.alpha.detach().item(),
        }

        if "per_weights" in batch:
            info.update({"errors": errors})

        grads = {"q_grads": q_grads, "pi_grads": pi_grads}

        info.update(addons_info)

        return grads, info

    def update_target_networks(self):
        """Update actor critic target networks with polyak averaging"""
        if self.iter % self.target_update_interval == 0:
            with torch.no_grad():
                for p, p_targ in zip(self.actor.parameters(), self.actor_targ.parameters()):
                    p_targ.data.mul_(self.polyak)
                    p_targ.data.add_((1 - self.polyak) * p.data)

    def apply_gradients(self, gradients=None):
        """
        Take an optimization step, previously setting new gradients if provided.
        Update also target networks.

        Parameters
        ----------
        gradients : list of tensors
            List of actor gradients.
        """
        if gradients is not None:
            set_gradients(
                self.actor.policy_net,
                gradients=gradients["pi_grads"], device=self.device)
            set_gradients(
                self.actor.q1, self.actor.q2,
                gradients=gradients["q_grads"], device=self.device)

        self.q_optimizer.step()
        self.pi_optimizer.step()
        self.alpha_optimizer.step()
        self.alpha = self.log_alpha.detach().exp()

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def set_weights(self, actor_weights):
        """
        Update actor with the given weights. Update also target networks.

        Parameters
        ----------
        actor_weights : dict of tensors
            Dict containing actor weights to be set.
        """
        self.actor.load_state_dict(actor_weights)
        self.alpha_optimizer.step()
        self.alpha = self.log_alpha.detach().exp()

        # Update target networks by polyak averaging.
        self.iter += 1
        self.update_target_networks()

    def update_algorithm_parameter(self, parameter_name, new_parameter_value):
        """
        If `parameter_name` is an attribute of the algorithm, change its value
        to `new_parameter_value value`.

        Parameters
        ----------
        parameter_name : str
            Worker.algo attribute name
        new_parameter_value : int or float
            New value for `parameter_name`.
        """
        if hasattr(self, parameter_name):
            setattr(self, parameter_name, new_parameter_value)
        if parameter_name == "lr":
            for param_group in self.pi_optimizer.param_groups:
                param_group['lr'] = new_parameter_value
            for param_group in self.q_optimizer.param_groups:
                param_group['lr'] = new_parameter_value
            for param_group in self.alpha_optimizer.param_groups:
                param_group['lr'] = new_parameter_value



import csv
import time
import json
from glob import glob
import os.path as osp
from gym.core import Wrapper


class Monitor(Wrapper):
    EXT = "monitor.csv"
    f = None

    def __init__(self, env, filename, allow_early_resets=False, reset_keywords=(), info_keywords=()):
        Wrapper.__init__(self, env=env)
        self.tstart = time.time()
        if filename:
            self.results_writer = ResultsWriter(
                filename,
                header={"t_start": time.time(), 'env_id': env.spec and env.spec.id},
                extra_keys=reset_keywords + info_keywords)
        else:
            self.results_writer = None
        self.reset_keywords = reset_keywords
        self.info_keywords = info_keywords
        self.allow_early_resets = allow_early_resets
        self.rewards = None
        self.needs_reset = True
        self.episode_rewards = []
        self.episode_lengths = []
        self.episode_times = []
        self.total_steps = 0
        self.current_reset_info = {} # extra info about the current episode, that was passed in during reset()

    def reset(self, **kwargs):
        self.reset_state()
        for k in self.reset_keywords:
            v = kwargs.get(k)
            if v is None:
                raise ValueError('Expected you to pass kwarg %s into reset'%k)
            self.current_reset_info[k] = v
        return self.env.reset(**kwargs)

    def reset_state(self):
        if not self.allow_early_resets and not self.needs_reset:
            raise RuntimeError("Tried to reset an environment before done. If you want to allow early resets, wrap your env with Monitor(env, path, allow_early_resets=True)")
        self.rewards = []
        self.needs_reset = False


    def step(self, action):
        if self.needs_reset:
            raise RuntimeError("Tried to step environment that needs reset")
        ob, rew, done, info = self.env.step(action)
        self.update(ob, rew, done, info)
        return (ob, rew, done, info)

    def update(self, ob, rew, done, info):
        self.rewards.append(rew)
        if done:
            self.needs_reset = True
            eprew = sum(self.rewards)
            eplen = len(self.rewards)
            epinfo = {"r": round(eprew, 6), "l": eplen, "t": round(time.time() - self.tstart, 6)}
            for k in self.info_keywords:
                epinfo[k] = info[k]
            self.episode_rewards.append(eprew)
            self.episode_lengths.append(eplen)
            self.episode_times.append(time.time() - self.tstart)
            epinfo.update(self.current_reset_info)
            if self.results_writer:
                self.results_writer.write_row(epinfo)
            assert isinstance(info, dict)
            if isinstance(info, dict):
                info['episode'] = epinfo

        self.total_steps += 1

    def close(self):
        super(Monitor, self).close()
        if self.f is not None:
            self.f.close()

    def get_total_steps(self):
        return self.total_steps

    def get_episode_rewards(self):
        return self.episode_rewards

    def get_episode_lengths(self):
        return self.episode_lengths

    def get_episode_times(self):
        return self.episode_times

class LoadMonitorResultsError(Exception):
    pass


class ResultsWriter(object):
    def __init__(self, filename, header='', extra_keys=()):
        self.extra_keys = extra_keys
        assert filename is not None
        if not filename.endswith(Monitor.EXT):
            if osp.isdir(filename):
                filename = osp.join(filename, Monitor.EXT)
            else:
                filename = filename + "." + Monitor.EXT

        already_exists = osp.isfile(filename)
        self.f = open(filename, "a+")
        if not already_exists:
            if isinstance(header, dict):
                header = '# {} \n'.format(json.dumps(header))
            self.f.write(header)
        self.f.flush()
        self.logger = csv.DictWriter(self.f, fieldnames=('r', 'l', 't')+tuple(extra_keys))
        if not already_exists:
            self.logger.writeheader()
        self.f.flush()

    def write_row(self, epinfo):
        if self.logger:
            self.logger.writerow(epinfo)
            self.f.flush()


def get_monitor_files(dir):
    return glob(osp.join(dir, "*" + Monitor.EXT))


def load_results(dir):
    import pandas
    monitor_files = (
        glob(osp.join(dir, "*monitor.json")) +
        glob(osp.join(dir, "*monitor.csv")))  # get both csv and (old) json files
    if not monitor_files:
        raise LoadMonitorResultsError("no monitor files of the form *%s found in %s" % (Monitor.EXT, dir))
    dfs = []
    headers = []
    for fname in monitor_files:
        with open(fname, 'rt') as fh:
            if fname.endswith('csv'):
                firstline = fh.readline()
                if not firstline:
                    continue
                assert firstline[0] == '#'
                header = json.loads(firstline[1:])
                df = pandas.read_csv(fh, index_col=None)
                headers.append(header)
            elif fname.endswith('json'):  # Deprecated json format
                episodes = []
                lines = fh.readlines()
                header = json.loads(lines[0])
                headers.append(header)
                for line in lines[1:]:
                    episode = json.loads(line)
                    episodes.append(episode)
                df = pandas.DataFrame(episodes)
            else:
                assert 0, 'unreachable'
            df['t'] += header['t_start']
        dfs.append(df)
    df = pandas.concat(dfs)
    df.sort_values('t', inplace=True)
    df.reset_index(inplace=True)
    df['t'] -= min(header['t_start'] for header in headers)
    df.headers = headers  # HACK to preserve backwards compatibility
    return df



# coding=utf-8
"""
Vocabulary helper class

from https://github.com/MolecularAI/reinvent-models/blob/main/reinvent_models/reinvent_core/models/vocabulary.py

str --> list of str --> np.array
str --> list of str: tokenizer.tokenize
list of str --> np.array: vocabulary.encode

"""

import re
import numpy as np


# contains the data structure
class Vocabulary:
    """Stores the tokens and allows their conversion to vocabulary indexes."""

    def __init__(self, tokens=None, starting_id=0):
        self._tokens = {}
        self._current_id = starting_id
        if tokens:
            for token, idx in tokens.items():
                self._add(token, idx)
                self._current_id = max(self._current_id, idx + 1)

    @property
    def vocab_size(self):
        """Vocabulary size"""
        return len(self._tokens) // 2

    def __getitem__(self, token_or_id):
        return self._tokens[token_or_id]

    def add(self, token):
        """Adds a token."""
        if not isinstance(token, str):
            raise TypeError("Token is not a string")
        if token in self:
            return self[token]
        self._add(token, self._current_id)
        self._current_id += 1
        return self._current_id - 1

    def update(self, tokens):
        """Adds many tokens."""
        return [self.add(token) for token in tokens]

    def __delitem__(self, token_or_id):
        other_val = self._tokens[token_or_id]
        del self._tokens[other_val]
        del self._tokens[token_or_id]

    def __contains__(self, token_or_id):
        return token_or_id in self._tokens

    def __len__(self):
        return len(self._tokens) // 2

    def encode(self, tokens):
        """Encodes a list of tokens as vocabulary indexes."""
        vocab_index = np.zeros(len(tokens), dtype=np.float32)
        for i, token in enumerate(tokens):
            vocab_index[i] = self._tokens[token]
        return vocab_index

    def decode(self, vocab_index):
        """Decodes a vocabulary index matrix to a list of tokens."""
        tokens = []
        for idx in vocab_index:
            tokens.append(self[idx])
        return tokens

    def _add(self, token, idx):
        if idx not in self._tokens:
            self._tokens[token] = idx
            self._tokens[idx] = token
        else:
            raise ValueError("IDX already present in vocabulary")

    def tokens(self):
        """Returns the tokens from the vocabulary"""
        return [t for t in self._tokens if isinstance(t, str)]


class SMILESTokenizer:
    """Deals with the tokenization and untokenization of SMILES."""

    REGEXPS = {
        "brackets": re.compile(r"(\[[^\]]*\])"),
        "2_ring_nums": re.compile(r"(%\d{2})"),
        "brcl": re.compile(r"(Br|Cl)")
    }
    REGEXP_ORDER = ["brackets", "2_ring_nums", "brcl"]

    def tokenize(self, data, with_begin_and_end=True):
        """Tokenizes a SMILES string."""
        def split_by(data, regexps):
            if not regexps:
                return list(data)
            regexp = self.REGEXPS[regexps[0]]
            splitted = regexp.split(data)
            tokens = []
            for i, split in enumerate(splitted):
                if i % 2 == 0:
                    tokens += split_by(split, regexps[1:])
                else:
                    tokens.append(split)
            return tokens

        tokens = split_by(data, self.REGEXP_ORDER)
        if with_begin_and_end:
            tokens = ["^"] + tokens + ["$"]
        return tokens

    def untokenize(self, tokens):
        """Untokenizes a SMILES string."""
        smi = ""
        for token in tokens:
            if token == "$":
                break
            if token != "^":
                smi += token
        return smi


def create_vocabulary(smiles_list, tokenizer):
    """Creates a vocabulary for the SMILES syntax."""
    tokens = set()
    for smi in smiles_list:
        tokens.update(tokenizer.tokenize(smi, with_begin_and_end=False))

    vocabulary = Vocabulary()
    vocabulary.update(["<pad>", "$", "^"] + sorted(tokens))
    return vocabulary


def create_vocabularies(scaffold_list, scaffold_tokenizer, decoration_list, decoration_tokenizer):
    """Creates a vocabulary for the SMILES syntax."""
    scaffold_tokens = set()
    decoration_tokens = set()
    for scaffold, decoration in zip(scaffold_list, decoration_list):
        scaffold_tokens.update(scaffold_tokenizer.tokenize(scaffold, with_begin_and_end=False))
        decoration_tokens.update(decoration_tokenizer.tokenize(decoration, with_begin_and_end=False))

    scaffold_vocabulary = Vocabulary()
    # vocabulary.update(["$", "^"] + sorted(tokens))  # end token is 0 (also counts as padding)
    scaffold_vocabulary.update(["<pad>", "$", "^"] + sorted(scaffold_tokens))

    decoration_vocabulary = Vocabulary()
    decoration_vocabulary.update(["<pad>", "$", "^"] + sorted(decoration_tokens))

    return scaffold_vocabulary, decoration_vocabulary


class ReinventVocabulary:

    def __init__(self, vocabulary, tokenizer):
        self.vocabulary = vocabulary
        self.tokenizer = tokenizer

    def encode_smile(self, smile, with_begin_and_end=True):
        """Encodes a SMILE from str to np.array."""
        return self.vocabulary.encode(self.tokenizer.tokenize(smile, with_begin_and_end))

    def decode_smile(self, encoded_smile):
        """Decodes a SMILE from np.array to str."""
        return self.tokenizer.untokenize(self.vocabulary.decode(encoded_smile))

    def encode_token(self, token):
        """Encodes token from str to int"""
        return self.vocabulary.encode([str(token)])[0]

    def decode_token(self, token):
        """Decodes token from int to str"""
        return self.vocabulary.decode([int(token)])[0]

    def remove_start_and_end_tokens(self, smile):
        """Remove start and end tokens from a SMILE"""
        return self.tokenizer.untokenize(smile)

    def count_tokens(self, smile):
        return len(self.tokenizer.tokenize(smile))

    def __len__(self):
        """Returns the length of the vocabulary."""
        return len(self.vocabulary)

    @classmethod
    def from_list(cls, smiles_list):
        """Creates the vocabulary from a list of smiles."""
        tokenizer = SMILESTokenizer()
        vocabulary = create_vocabulary(smiles_list, tokenizer)
        return ReinventVocabulary(vocabulary, tokenizer)


class LibinventVocabulary:

    def __init__(self, scaffold_vocabulary, scaffold_tokenizer, decoration_vocabulary, decoration_tokenizer):
        self.scaffold_vocabulary = scaffold_vocabulary
        self.scaffold_tokenizer = scaffold_tokenizer
        self.decoration_vocabulary = decoration_vocabulary
        self.decoration_tokenizer = decoration_tokenizer

    def len_scaffold(self):
        """Returns the length of the scaffold vocabulary."""
        return len(self.scaffold_vocabulary)

    def len_decoration(self):
        """Returns the length of the decoration vocabulary."""
        return len(self.decoration_vocabulary)

    def encode_scaffold(self, scaffold):
        """Encodes a scaffold from str to np.array."""
        return self.scaffold_vocabulary.encode(self.scaffold_tokenizer.tokenize(scaffold))

    def decode_scaffold(self, encoded_scaffold):
        """Decodes the scaffold from np.array to str."""
        return self.scaffold_tokenizer.untokenize(self.scaffold_vocabulary.decode(encoded_scaffold))

    def encode_scaffold_token(self, token):
        """Encodes token from str to int"""
        return self.scaffold_vocabulary.encode([str(token)])[0]

    def decode_scaffold_token(self, token):
        """Decodes token from int to str"""
        return self.scaffold_vocabulary.decode([int(token)])[0]

    def encode_decoration(self, smiles):
        """Encodes a decoration from str to np.array."""
        return self.decoration_vocabulary.encode(self.decoration_tokenizer.tokenize(smiles))

    def decode_decoration(self, encoded_decoration):
        """Decodes a decoration from np.array to str."""
        return self.decoration_tokenizer.untokenize(self.decoration_vocabulary.decode(encoded_decoration))

    def encode_decoration_token(self, token):
        """Encodes token from str to int"""
        return self.decoration_vocabulary.encode([str(token)])[0]

    def decode_decoration_token(self, token):
        """Decodes token from int to str"""
        return self.decoration_vocabulary.decode([int(token)])[0]

    def count_scaffold_tokens(self, scaffold):
        return len(self.scaffold_tokenizer.tokenize(scaffold))

    def remove_start_and_end_tokens(self, smile):
        """Remove start and end tokens from a SMILE"""
        return self.decoration_tokenizer.untokenize(smile)

    @classmethod
    def from_lists(cls, scaffold_list, decoration_list):
        """Creates the vocabularies from lists."""
        scaffold_tokenizer, decoration_tokenizer = SMILESTokenizer(), SMILESTokenizer()
        scaffold_vocabulary, decoration_vocabulary = create_vocabularies(
            scaffold_list, scaffold_tokenizer, decoration_list, decoration_tokenizer)
        return LibinventVocabulary(scaffold_vocabulary, scaffold_tokenizer, decoration_vocabulary, decoration_tokenizer)





import pandas as pd


class DiversityFilterMemory:

    def __init__(self):
        df_dict = {"Step": [], "SMILES": []}
        self._memory_dataframe = pd.DataFrame(df_dict)

    def update(self, smile: str):
        if not self.smiles_exists(smile):
            self._add_to_memory_dataframe(smile)

    def _add_to_memory_dataframe(self, smile: str):
        data, headers = [], []
        headers.append("SMILES")
        data.append(smile)
        new_data = pd.DataFrame([data], columns=headers)
        self._memory_dataframe = pd.concat([self._memory_dataframe, new_data], ignore_index=True, sort=False)

    def get_memory(self) -> pd.DataFrame:
        return self._memory_dataframe

    def set_memory(self, memory: pd.DataFrame):
        self._memory_dataframe = memory

    def smiles_exists(self, smiles: str):
        if len(self._memory_dataframe) == 0:
            return False
        return smiles in self._memory_dataframe['SMILES'].values

    def number_of_smiles(self):
        return len(set(self._memory_dataframe["SMILES"].values))




import gym
import math
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


class FixupCNN(nn.Module):
    """
    A larger version of the IMPALA CNN with Fixup init.
    See Fixup: https://arxiv.org/abs/1901.09321.
    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    rgb_norm : bool
        Whether or not to divide input by 255.
    """
    def __init__(self, input_space, output_size=256, rgb_norm=True, activation=nn.ReLU, final_activation=True):

        super(FixupCNN, self).__init__()

        self.rgb_norm = rgb_norm

        if isinstance(input_space, gym.Space):
            input_shape = input_space.shape
        else:
            input_shape = input_space

        if len(input_shape) != 3:
            raise ValueError("Trying to extract features with Fixup CNN for "
                             "an obs space with len(shape) != 3")

        depth_in = input_shape[0]
        layers = []
        for depth_out in [16, 32, 32]:
            layers.extend([
                nn.Conv2d(depth_in, depth_out, 3, padding=1),
                nn.MaxPool2d(3, stride=2, padding=1),
                FixupResidualModule(depth_out, 6),
                FixupResidualModule(depth_out, 6),
            ])
            depth_in = depth_out
        self.feature_extractor = nn.Sequential(*layers)

        # Define final layer
        feature_size = int(np.prod(self.feature_extractor(
            torch.randn(1, *input_space.shape)).shape))
        head = [nn.Linear(feature_size, output_size)]
        if final_activation:
            head += [nn.ReLU(inplace=True)]
        self.head = nn.Sequential(*head)

        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network
        Parameters
        ----------
        inputs : torch.tensor
            Input data.
        Returns
        -------
        out : torch.tensor
            Output feature map.
        """

        if self.rgb_norm:
            inputs = inputs / 255.0

        out = self.feature_extractor(inputs).view(inputs.size(0), -1)
        return out


class FixupResidualModule(nn.Module):
    """
    FixupCNN resudial block.
    Parameters
    ----------
    depth : int
        Number of input feature maps.
    num_residual : int
        Number of residual feature blocks in the architecture
    """
    def __init__(self, depth, num_residual):
        super(FixupResidualModule, self).__init__()

        self.conv1 = nn.Conv2d(depth, depth, 3, padding=1, bias=False)
        self.conv2 = nn.Conv2d(depth, depth, 3, padding=1, bias=False)

        for p in self.conv1.parameters():
            p.data.mul_(1 / math.sqrt(num_residual))
        for p in self.conv2.parameters():
            p.data.zero_()

        self.bias1 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias2 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias3 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.bias4 = nn.Parameter(torch.zeros([depth, 1, 1]))
        self.scale = nn.Parameter(torch.ones([depth, 1, 1]))

    def forward(self, x):
        """
        Forward pass residual block
        Parameters
        ----------
        inputs : torch.tensor
            Input feature map.
        Returns
        -------
        out : torch.tensor
            Output feature map.
        """
        x = F.relu(x)
        out = x + self.bias1
        out = self.conv1(out)
        out = out + self.bias2
        out = F.relu(out)
        out = out + self.bias3
        out = self.conv2(out)
        out = out * self.scale
        out = out + self.bias4
        return out + x



import multiprocessing as mp

import numpy as np
from pytorchrl.agent.env.openai_baselines_dependencies.vec_envs.vec_env_base import VecEnvBase, CloudpickleWrapper,\
    clear_mpi_env_vars


def worker(remote, parent_remote, env_fn_wrappers):
    def step_env(env, action):
        ob, reward, done, info = env.step(action)
        if done:
            ob = env.reset()
        return ob, reward, done, info

    parent_remote.close()
    envs = [env_fn_wrapper() for env_fn_wrapper in env_fn_wrappers.x]
    try:
        while True:
            cmd, data = remote.recv()
            if cmd == 'step':
                remote.send([step_env(env, action) for env, action in zip(envs, data)])
            elif cmd == 'reset':
                remote.send([env.reset() for env in envs])
            elif cmd == 'render':
                remote.send([env.render(mode='rgb_array') for env in envs])
            elif cmd == 'close':
                remote.close()
                break
            elif cmd == 'get_spaces_spec':
                remote.send(CloudpickleWrapper((envs[0].observation_space, envs[0].action_space, envs[0].spec)))
            else:
                raise NotImplementedError
    except KeyboardInterrupt:
        print('SubprocVecEnv worker: got KeyboardInterrupt')
    finally:
        for env in envs:
            env.close()


class SubprocVecEnv(VecEnvBase):
    """
    VecEnv that runs multiple environments in parallel in subproceses and communicates with them via pipes.
    Recommended to use when num_envs > 1 and step() can be a bottleneck.
    """
    def __init__(self, env_fns, spaces=None, context='spawn', in_series=1):
        """
        Arguments:
        env_fns: iterable of callables -  functions that create environments to run in subprocesses. Need to be cloud-pickleable
        in_series: number of environments to run in series in a single process
        (e.g. when len(env_fns) == 12 and in_series == 3, it will run 4 processes, each running 3 envs in series)
        """
        self.waiting = False
        self.closed = False
        self.in_series = in_series
        nenvs = len(env_fns)
        assert nenvs % in_series == 0, "Number of envs must be divisible by number of envs to run in series"
        self.nremotes = nenvs // in_series
        env_fns = np.array_split(env_fns, self.nremotes)
        ctx = mp.get_context(context)
        self.remotes, self.work_remotes = zip(*[ctx.Pipe() for _ in range(self.nremotes)])
        self.ps = [ctx.Process(target=worker, args=(work_remote, remote, CloudpickleWrapper(env_fn)))
                   for (work_remote, remote, env_fn) in zip(self.work_remotes, self.remotes, env_fns)]
        for p in self.ps:
            p.daemon = True  # if the main process crashes, we should not cause things to hang
            with clear_mpi_env_vars():
                p.start()
        for remote in self.work_remotes:
            remote.close()

        self.remotes[0].send(('get_spaces_spec', None))
        observation_space, action_space, self.spec = self.remotes[0].recv().x
        self.viewer = None
        VecEnvBase.__init__(self, nenvs, observation_space, action_space)

    def step_async(self, actions):
        self._assert_not_closed()
        actions = np.array_split(actions, self.nremotes)
        for remote, action in zip(self.remotes, actions):
            remote.send(('step', action))
        self.waiting = True

    def step_wait(self):
        self._assert_not_closed()
        results = [remote.recv() for remote in self.remotes]
        results = _flatten_list(results)
        self.waiting = False
        obs, rews, dones, infos = zip(*results)
        return _flatten_obs(obs), np.stack(rews), np.stack(dones), infos

    def reset(self):
        self._assert_not_closed()
        for remote in self.remotes:
            remote.send(('reset', None))
        obs = [remote.recv() for remote in self.remotes]
        obs = _flatten_list(obs)
        return _flatten_obs(obs)

    def reset_single_env(self, num_env):
        self._assert_not_closed()
        self.remotes[num_env].send(('reset', None))
        obs = self.remotes[num_env].recv()
        return _flatten_obs(obs)

    def close_extras(self):
        self.closed = True
        if self.waiting:
            for remote in self.remotes:
                remote.recv()
        for remote in self.remotes:
            remote.send(('close', None))
        for p in self.ps:
            p.join()

    def get_images(self):
        self._assert_not_closed()
        for pipe in self.remotes:
            pipe.send(('render', None))
        imgs = [pipe.recv() for pipe in self.remotes]
        imgs = _flatten_list(imgs)
        return imgs

    def _assert_not_closed(self):
        assert not self.closed, "Trying to operate on a SubprocVecEnv after calling close()"

    def __del__(self):
        if not self.closed:
            try:
                self.close()
            except Exception:
                pass


def _flatten_obs(obs):
    assert isinstance(obs, (list, tuple))
    assert len(obs) > 0

    if isinstance(obs[0], dict):
        keys = obs[0].keys()
        return {k: np.stack([o[k] for o in obs]) for k in keys}
    else:
        return np.stack(obs)


def _flatten_list(l):
    assert isinstance(l, (list, tuple))
    assert len(l) > 0
    assert all([len(l_) > 0 for l_ in l])

    return [l__ for l_ in l for l__ in l_]


import gym
import numpy as np
import torch
import torch.nn as nn

from pytorchrl.agent.actors.utils import init
from pytorchrl.agent.actors.feature_extractors.mlp import MLP
from pytorchrl.agent.actors.feature_extractors.cnn import CNN


class DictNet(nn.Module):
    """
    Neural network architecture designed to deal with dict observation spaces.
    It extracts individual features independently with either a MLP or a CNN,
    and then combines the information with a final sequence of fully connected
    layers.

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    shared_hidden_sizes : list
        Hidden layers sizes used to combine multimodel features.
    activation : func
        Non-linear activation function.
    """
    def __init__(self, input_space, shared_hidden_sizes=[256, 256], output_size=256, activation=nn.ReLU):
        super(DictNet, self).__init__()

        assert isinstance(input_space, gym.spaces.Dict)

        init_ = lambda m: init(
            m, nn.init.orthogonal_, lambda x: nn.init. constant_(
                x, 0), nn.init.calculate_gain('relu'))

        self.net_branches = nn.ModuleDict()
        for k in input_space.spaces:
            if len(input_space[k].shape) <= 2:
                net = MLP(input_space[k], activation=activation)
            elif len(input_space[k].shape) == 3:
                net = CNN(input_space[k], activation=activation)
            else:
                raise ValueError(
                    "Observation space with nested dicts not allowed.")
            self.net_branches[k] = net

        # Get concat output map size
        map_size = 0
        sample_obs = input_space.sample()
        for k in sample_obs:
            map_size += int(np.prod(
                self.net_branches[k](torch.from_numpy(
                    sample_obs[k]).float().unsqueeze(0)).shape))

        # Define shared layers
        layers = []
        sizes = [map_size] + shared_hidden_sizes + [output_size]
        for j in range(len(sizes) - 1):
            layers += [init_(nn.Linear(sizes[j], sizes[j + 1])), activation()]
        self.shared_net = nn.Sequential(*layers)

        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """

        assert isinstance(inputs, dict)

        output = []
        for k in inputs:
            x = inputs[k].view(inputs[k].size(0), -1)
            output.append(self.net_branches[k](x))

        output = torch.cat(output, dim=1)

        return self.shared_net(output)



import gym
import numpy as np
import torch
import torch.nn as nn
from pytorchrl.agent.actors.utils import init
from pytorchrl.agent.actors.feature_extractors.utils import get_gain


class CNN(nn.Module):
    """
    Convolutional Neural Network.

    Parameters
    ----------
    input_space : gym.Space
        Environment observation space.
    rgb_norm : bool
        Whether or not to divide input by 255.
    activation : func
        Non-linear activation function.
    final_activation : bool
        Whether or not to apply activation function after last layer.
    strides : list
        Convolutional layers strides.
    filters : list
        Convolutional layers number of filters.
    kernel_sizes : list
        Convolutional layers kernel sizes.
    output_sizes : list
        output hidden layers sizes.
    """
    def __init__(self,
                 input_space,
                 rgb_norm=True,
                 activation=nn.ReLU,
                 final_activation=True,
                 strides=[4, 2, 1],
                 filters=[32, 64, 64],
                 kernel_sizes=[8, 4, 3],
                 output_sizes=[256, 448]):

        super(CNN, self).__init__()

        self.rgb_norm = rgb_norm

        assert len(filters) == len(strides) and len(strides) == len(kernel_sizes)

        if isinstance(input_space, gym.Space):
            input_shape = input_space.shape
        else:
            input_shape = input_space

        if len(input_shape) != 3:
            raise ValueError("Trying to extract features with a CNN for an obs space with len(shape) != 3")

        # Define CNN feature extractor
        layers = []
        filters = [input_shape[0]] + filters
        for j in range(len(filters) - 1):
            layers += [nn.Conv2d(
                filters[j], filters[j + 1], stride=strides[j],
                kernel_size=kernel_sizes[j]), activation()]
        self.feature_extractor = nn.Sequential(*layers)

        # Define final MLP layers
        feature_size = int(np.prod(self.feature_extractor(torch.randn(1, *input_shape)).shape))
        layers = []
        sizes = [feature_size] + output_sizes
        for j in range(len(sizes) - 1):
            layers += [nn.Linear(sizes[j], sizes[j + 1])]
            if j < len(sizes) - 2 or final_activation:
                layers += [activation()]
        self.head = nn.Sequential(*layers)

        for layer in self.feature_extractor.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.orthogonal_(layer.weight, gain=get_gain(activation))
                layer.bias.data.zero_()
            elif isinstance(layer, nn.Linear):
                nn.init.orthogonal_(layer.weight, gain=get_gain(activation))
                layer.bias.data.zero_()

        for layer in self.head.modules():
            if isinstance(layer, nn.Conv2d):
                nn.init.orthogonal_(layer.weight, gain=get_gain(activation))
                layer.bias.data.zero_()
            elif isinstance(layer, nn.Linear):
                nn.init.orthogonal_(layer.weight, gain=get_gain(activation))
                layer.bias.data.zero_()

        self.train()

    def forward(self, inputs):
        """
        Forward pass Neural Network

        Parameters
        ----------
        inputs : torch.tensor
            Input data.

        Returns
        -------
        out : torch.tensor
            Output feature map.
        """

        if self.rgb_norm:
            inputs = inputs / 255.0

        out = self.feature_extractor(inputs)
        out = out.contiguous()
        out = out.view(inputs.size(0), -1)
        out = self.head(out)

        return out



import torch


class StandardScaler(object):

    def __init__(self, device):
        self.input_mu = torch.zeros(1).to(device)
        self.input_std = torch.ones(1).to(device)
        self.target_mu = torch.zeros(1).to(device)
        self.target_std = torch.ones(1).to(device)
        self.device = device

    def fit(self, inputs, targets):
        """
        Runs two ops, one for assigning the mean of the data to the internal mean, and
        another for assigning the standard deviation of the data to the internal standard deviation.
        This function must be called within a 'with <session>.as_default()' block.

        Parameters
        ----------
        inputs : torch.Tensor
            A torch Tensor containing the input
        targets : torch.Tensor
            A torch Tensor containing the input

        """
        self.input_mu = torch.mean(inputs, dim=0, keepdims=True).to(self.device)
        self.input_std = torch.std(inputs, dim=0, keepdims=True).to(self.device)
        self.input_std[self.input_std < 1e-8] = 1.0
        self.target_mu = torch.mean(targets, dim=0, keepdims=True).to(self.device)
        self.target_std = torch.std(targets, dim=0, keepdims=True).to(self.device)
        self.target_std[self.target_std < 1e-8] = 1.0

    def transform(self, inputs, targets=None):
        """
        Transforms the input matrix data using the parameters of this scaler.

        Parameters
        ----------
        inputs : torch.Tensor
            A torch Tensor containing the points to be transformed.
        targets : torch.Tensor
            A torch Tensor containing the points to be transformed.

        Returns
        -------
        norm_inputs : torch.Tensor
            Normalized inputs
        norm_targets : torch.Tensor
            Normalized targets
        """
        norm_inputs = (inputs - self.input_mu) / self.input_std
        norm_targets = None
        if targets is not None:
            norm_targets = (targets - self.target_mu) / self.target_std
        return norm_inputs, norm_targets

    def inverse_transform(self, targets):
        """
        Undoes the transformation performed by this scaler.

        Parameters
        ----------
        targets : torch.Tensor
            A torch Tensor containing the points to be transformed.

        Returns
        -------
        output : torch.Tensor
            The transformed dataset.
        """
        output = self.target_std * targets + self.target_mu
        return output



import torch
import pytorchrl as prl
from pytorchrl.agent.storages.on_policy.vanilla_on_policy_buffer import VanillaOnPolicyBuffer as B


class GAEBuffer(B):
    """
    Storage class for On-Policy algorithms with Generalized Advantage
    Estimator (GAE). https://arxiv.org/abs/1506.02438

    Parameters
    ----------
    size : int
        Storage capacity along time axis.
    gae_lambda : float
        GAE lambda parameter.
    device: torch.device
        CPU or specific GPU where data tensors will be placed and class
        computations will take place. Should be the same device where the
        actor model is located.
    envs : VecEnv
        Vector of environments instance.
    actor : Actor
        Actor class instance.
    algorithm : Algorithm
        Algorithm class instance.
    """

    # Data fields to store in buffer and contained in generated batches
    storage_tensors = prl.OnPolicyDataKeys

    def __init__(self, size, device, actor, algorithm, envs, gae_lambda=0.95):

        super(GAEBuffer, self).__init__(
            size=size,
            envs=envs,
            device=device,
            actor=actor,
            algorithm=algorithm)

        self.gae_lambda = gae_lambda

    @classmethod
    def create_factory(cls, size, gae_lambda=0.95):
        """
        Returns a function that creates OnPolicyGAEBuffer instances.

        Parameters
        ----------
        size : int
            Storage capacity along time axis.
        gae_lambda : float
            GAE lambda parameter.

        Returns
        -------
        create_buffer_instance : func
            creates a new OnPolicyBuffer class instance.
        """
        def create_buffer_instance(device, actor, algorithm, envs):
            """Create and return a OnPolicyGAEBuffer instance."""
            return cls(size, device, actor, algorithm, envs, gae_lambda)
        return create_buffer_instance

    @property
    def used_capacity(self):
        """Returns the step up to which storage is full with env transitions."""
        return self.step - 1 if self.step != 0 else self.max_size

    def compute_returns(self, rewards, returns, values, dones, gamma):
        """Compute return values."""
        gae_lambda, length, gae = self.gae_lambda, self.used_capacity, 0
        returns[length].copy_(values[length])
        for step in reversed(range(length)):
            delta = (rewards[step] + gamma * values[step + 1] * (1.0 - dones[step + 1]) - values[step])
            gae = delta + gamma * gae_lambda * (1.0 - dones[step + 1]) * gae
            returns[step] = gae + values[step]





"""wrappers from https://github.com/openai/baselines/blob/master/baselines/common/atari_wrappers.py"""

import os

os.environ.setdefault('PATH', '')

import gym
import cv2
import numpy as np
from collections import deque

cv2.ocl.setUseOpenCL(False)
from pytorchrl import EMBED
from pytorchrl.envs.common import FrameStack
from pytorchrl.envs.atari.utils import imdownscale


class StickyActionEnv(gym.Wrapper):
    def __init__(self, env, p=0.25):
        super(StickyActionEnv, self).__init__(env)
        self.p = p
        self.last_action = 0

    def step(self, action):
        if np.random.uniform() < self.p:
            action = self.last_action

        self.last_action = action
        return self.env.step(action)

    def reset(self):
        self.last_action = 0
        return self.env.reset()


class TimeLimit(gym.Wrapper):
    def __init__(self, env, max_episode_steps=None):
        super(TimeLimit, self).__init__(env)
        self._max_episode_steps = max_episode_steps
        self._elapsed_steps = 0

    def step(self, ac):
        observation, reward, done, info = self.env.step(ac)
        self._elapsed_steps += 1
        if self._elapsed_steps >= self._max_episode_steps:
            done = True
            info['TimeLimit.truncated'] = True
        return observation, reward, done, info

    def reset(self, **kwargs):
        self._elapsed_steps = 0
        return self.env.reset(**kwargs)


class NoopResetEnv(gym.Wrapper):
    def __init__(self, env, noop_max=30):
        """Sample initial states by taking random number of no-ops on reset.
        No-op is assumed to be action 0.
        """
        gym.Wrapper.__init__(self, env)
        self.noop_max = noop_max
        self.override_num_noops = None
        self.noop_action = 0
        assert env.unwrapped.get_action_meanings()[0] == 'NOOP'

    def reset(self, **kwargs):
        """ Do no-op action for a number of steps in [1, noop_max]."""
        self.env.reset(**kwargs)
        if self.override_num_noops is not None:
            noops = self.override_num_noops
        else:
            noops = self.unwrapped.np_random.randint(1, self.noop_max + 1)  # pylint: disable=E1101
        assert noops > 0
        obs = None
        for _ in range(noops):
            obs, _, done, _ = self.env.step(self.noop_action)
            if done:
                obs = self.env.reset(**kwargs)
        return obs

    def step(self, ac):
        return self.env.step(ac)


class FireResetEnv(gym.Wrapper):
    def __init__(self, env):
        """Take action on reset for environments that are fixed until firing."""
        gym.Wrapper.__init__(self, env)
        assert env.unwrapped.get_action_meanings()[1] == 'FIRE'
        assert len(env.unwrapped.get_action_meanings()) >= 3

    def reset(self, **kwargs):
        self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(1)
        if done:
            self.env.reset(**kwargs)
        obs, _, done, _ = self.env.step(2)
        if done:
            self.env.reset(**kwargs)
        return obs

    def step(self, ac):
        return self.env.step(ac)


class EpisodicLifeEnv(gym.Wrapper):
    def __init__(self, env):
        """Make end-of-life == end-of-episode, but only reset on true game over.
        Done by DeepMind for the DQN and co. since it helps value estimation.
        """
        gym.Wrapper.__init__(self, env)
        self.lives = 0
        self.was_real_done = True
        self.total_reward = 0.0

    def step(self, action):
        obs, reward, done, info = self.env.step(action)

        self.total_reward += reward
        info['EpisodicReward'] = self.total_reward

        self.was_real_done = done
        # check current lives, make loss of life terminal,
        # then update lives to handle bonus lives
        lives = self.env.unwrapped.ale.lives()
        if lives < self.lives and lives > 0:
            # for Qbert sometimes we stay in lives == 0 condition for a few frames
            # so it's important to keep lives > 0, so that we only reset once
            # the environment advertises done.
            done = True
        info['Lives'] = lives
        self.lives = lives
        return obs, reward, done, info

    def reset(self, **kwargs):
        """Reset only when lives are exhausted.
        This way all states are still reachable even though lives are episodic,
        and the learner need not know about any of this behind-the-scenes.
        """
        if self.was_real_done:
            self.total_reward = 0.0
            obs = self.env.reset(**kwargs)
        else:
            # no-op step to advance from terminal/lost life state
            obs, _, _, _ = self.env.step(0)
        self.lives = self.env.unwrapped.ale.lives()
        return obs


class ClipRewardEnv(gym.Wrapper):
    def __init__(self, env):
        gym.Wrapper.__init__(self, env)
        self.total_reward = 0.0

    def step(self, action):
        obs, reward, done, info = self.env.step(action)

        self.total_reward += reward
        info['UnclippedReward'] = self.total_reward

        reward = np.sign(reward)
        return obs, reward, done, info

    def reset(self, **kwargs):
        self.total_reward = 0.0
        return self.env.reset(**kwargs)


class ScaleRewardEnv(gym.Wrapper):
    def __init__(self, env, scaling=0.001):
        gym.Wrapper.__init__(self, env)
        self.scaling = scaling
        self.total_reward = 0.0

    def step(self, action):
        obs, reward, done, info = self.env.step(action)
        self.total_reward += reward
        info['UnscaledReward'] = self.total_reward
        reward = self.scaling * reward
        return obs, reward, done, info

    def reset(self, **kwargs):
        self.total_reward = 0.0
        return self.env.reset(**kwargs)


class MaxAndSkipEnv(gym.Wrapper):
    def __init__(self, env, skip=4):
        """Return only every `skip`-th frame"""
        gym.Wrapper.__init__(self, env)
        # most recent raw observations (for max pooling across time steps)
        self._obs_buffer = np.zeros((2,) + env.observation_space.shape, dtype=np.uint8)
        self._skip = skip

    def step(self, action):
        """Repeat action, sum reward, and max over last observations."""
        total_reward = 0.0
        done = None
        for i in range(self._skip):
            obs, reward, done, info = self.env.step(action)
            if i == self._skip - 2:
                self._obs_buffer[0] = obs
            if i == self._skip - 1:
                self._obs_buffer[1] = obs
            total_reward += reward
            if done:
                break
        # Note that the observation on the done=True frame
        # doesn't matter
        max_frame = self._obs_buffer.max(axis=0)

        return max_frame, total_reward, done, info

    def reset(self, **kwargs):
        return self.env.reset(**kwargs)


class MontezumaVisitedRoomEnv(gym.Wrapper):
    def __init__(self, env, room_address):
        gym.Wrapper.__init__(self, env)
        self.room_address = room_address
        self.visited_rooms = set()  # Only stores unique numbers.

    def step(self, action):
        state, reward, done, info = self.env.step(action)
        ram = self.unwrapped.ale.getRAM()
        assert len(ram) == 128
        self.visited_rooms.add(ram[self.room_address])
        info['VisitedRooms'] = len(self.visited_rooms)
        return state, reward, done, info

    def reset(self):
        self.visited_rooms.clear()
        return self.env.reset()


class MontezumaEmbeddingsEnv(gym.Wrapper):
    def __init__(self, env, embeddings_shape=(11, 8), embeddings_num_values=8, use_domain_knowledge=False,
                 domain_knowledge_embedding="default", double_state=False):
        gym.Wrapper.__init__(self, env)

        # pos 0: The current frame of the episode.
        # pos 3: The current screen. Room?
        # pod 57: room level?
        # pos 19, 20, 21: The score, represented in Binary Coded Decimal. This is, every nibble represents a decimal digit
        # pos 42: joe_x, from 0 to 153
        # pos 43: joe_y, from 0 to 122 0 135 to 253
        # pos 52: agent orientation, 76 or 128
        # pos 65: inventory
        #       mallet +1
        #       key +2
        #       key +4
        #       key +8
        #       key +16
        #       sword +32
        #       sword +64
        #       torch +128

        self.joe_x = 42
        self.joe_y = 43
        self.room_level = 57
        self.room_address = 3
        self.joe_inventory = 65
        self.last_state = None
        self.embeddings_shape = embeddings_shape
        self.embeddings_num_values = embeddings_num_values
        self.use_domain_knowledge = use_domain_knowledge
        self.domain_knowledge_embedding = domain_knowledge_embedding
        self.double_state = double_state
        if double_state:
            self._embed_buff = deque(maxlen=2)

    def step(self, action):

        # Create embedding
        if self.use_domain_knowledge:
            ram = self.unwrapped.ale.getRAM()
            assert len(ram) == 128
            if self.domain_knowledge_embedding == "default":
                embed_state = np.array(
                    [
                        np.clip(ram[self.joe_x], 0, 153),  # range 0 - 153
                        np.clip(ram[self.joe_y], 135, 253),  # range 135 - 253
                        ram[self.room_address] + 24 * ram[self.room_level],
                        ram[self.joe_inventory],
                    ]
                )
            elif self.domain_knowledge_embedding == "room_inventory":
                embed_state = np.array(
                    [
                        ram[self.room_address] + 24 * ram[self.room_level],
                        ram[self.joe_inventory],
                    ]
                )
            elif self.domain_knowledge_embedding == "room":
                embed_state = np.array(
                    [
                        ram[self.room_address] + 24 * ram[self.room_level],
                    ]
                )
            state, reward, done, info = self.env.step(action)
        else:
            state, reward, done, info = self.env.step(action)
            embed_state = imdownscale(
                state=self.last_state[:, :, -1],
                target_shape=self.embeddings_shape,
                max_pix_value=self.embeddings_num_values)

        # Concat last 2 embeddings if specified
        if self.double_state:
            if len(self._embed_buff) < 2:
                self._embed_buff.append(embed_state)
                self._embed_buff.append(embed_state)
            if (embed_state != self._embed_buff[-1]).any():
                self._embed_buff.append(embed_state)
            embed_state = np.concatenate(self._embed_buff)

        info.update({EMBED: embed_state})
        self.last_state = state

        return state, reward, done, info

    def reset(self):
        self.last_state = self.env.reset()
        return self.last_state


class PitfallEmbeddingsEnv(gym.Wrapper):
    def __init__(self, env, embeddings_shape=(11, 8), embeddings_num_values=8, use_domain_knowledge=False,
                 double_state=False):
        gym.Wrapper.__init__(self, env)

        # pos 1: room ID
        self.room_level = 1
        self.last_state = None
        self.embeddings_shape = embeddings_shape
        self.embeddings_num_values = embeddings_num_values
        self.use_domain_knowledge = use_domain_knowledge
        self.double_state = double_state
        if double_state:
            self._embed_buff = deque(maxlen=2)

    def step(self, action):

        # Create embedding
        if self.use_domain_knowledge:
            ram = self.unwrapped.ale.getRAM()
            assert len(ram) == 128
            embed_state = np.array([np.clip(ram[self.room_level], 0, 255)])  # range 0 - 255
            state, reward, done, info = self.env.step(action)
        else:
            state, reward, done, info = self.env.step(action)
            embed_state = imdownscale(
                state=self.last_state[:, :, -1],
                target_shape=self.embeddings_shape,
                max_pix_value=self.embeddings_num_values)

        # Concat last 2 embeddings if specified
        if self.double_state:
            if len(self._embed_buff) < 2:
                self._embed_buff.append(embed_state)
                self._embed_buff.append(embed_state)
            if (embed_state != self._embed_buff[-1]).any():
                self._embed_buff.append(embed_state)
            embed_state = np.concatenate(self._embed_buff)

        info.update({EMBED: embed_state})
        self.last_state = state

        return state, reward, done, info

    def reset(self):
        self.last_state = self.env.reset()
        return self.last_state


class WarpFrame(gym.ObservationWrapper):
    def __init__(self, env, width=84, height=84, grayscale=True, dict_space_key=None):
        """
        Warp frames to 84x84 as done in the Nature paper and later work.
        If the environment uses dictionary observations, `dict_space_key` can be specified which indicates which
        observation should be warped.
        """
        super().__init__(env)
        self._width = width
        self._height = height
        self._grayscale = grayscale
        self._key = dict_space_key
        if self._grayscale:
            num_colors = 1
        else:
            num_colors = 3

        new_space = gym.spaces.Box(
            low=0,
            high=255,
            shape=(self._height, self._width, num_colors),
            dtype=np.uint8,
        )
        if self._key is None:
            original_space = self.observation_space
            self.observation_space = new_space
        else:
            original_space = self.observation_space.spaces[self._key]
            self.observation_space.spaces[self._key] = new_space
        assert original_space.dtype == np.uint8 and len(original_space.shape) == 3

    def observation(self, obs):
        if self._key is None:
            frame = obs
        else:
            frame = obs[self._key]

        if self._grayscale:
            frame = cv2.cvtColor(frame, cv2.COLOR_RGB2GRAY)
        frame = cv2.resize(
            frame, (self._width, self._height), interpolation=cv2.INTER_AREA
        )
        if self._grayscale:
            frame = np.expand_dims(frame, -1)

        if self._key is None:
            obs = frame
        else:
            obs = obs.copy()
            obs[self._key] = frame
        return obs


class ScaledFloatFrame(gym.ObservationWrapper):
    def __init__(self, env):
        gym.ObservationWrapper.__init__(self, env)
        self.observation_space = gym.spaces.Box(low=0, high=1, shape=env.observation_space.shape, dtype=np.float32)

    def observation(self, observation):
        # careful! This undoes the memory optimization, use
        # with smaller replay buffers only.
        return np.array(observation).astype(np.float32) / 255.0


class LazyFrames(object):
    def __init__(self, frames):
        """
        This object ensures that common frames between the observations are only stored once.
        It exists purely to optimize memory usage which can be huge for DQN's 1M frames replay
        buffers.
        This object should only be converted to numpy array before being passed to the model.
        You'd not believe how complex the previous solution was.
        """
        self._frames = frames
        self._out = None

    def _force(self):
        if self._out is None:
            self._out = np.concatenate(self._frames, axis=-1)
            self._frames = None
        return self._out

    def __array__(self, dtype=None):
        out = self._force()
        if dtype is not None:
            out = out.astype(dtype)
        return out

    def __len__(self):
        return len(self._force())

    def __getitem__(self, i):
        return self._force()[i]

    def count(self):
        frames = self._force()
        return frames.shape[frames.ndim - 1]

    def frame(self, i):
        return self._force()[..., i]


def wrap_deepmind(env, episode_life=True, clip_rewards=True, frame_stack=1, scale=False):
    """Configure environment for DeepMind-style Atari"""
    if episode_life:
        env = EpisodicLifeEnv(env)
    if 'FIRE' in env.unwrapped.get_action_meanings():
        env = FireResetEnv(env)
    env = WarpFrame(env)
    if scale:
        env = ScaledFloatFrame(env)
    if clip_rewards:
        env = ClipRewardEnv(env)
    env = FrameStack(env, frame_stack)
    return env


def make_atari(env_id, max_episode_steps=None, sticky_actions=False):
    env = gym.make(env_id)
    env._max_episode_steps = max_episode_steps * 4
    assert 'NoFrameskip' in env.spec.id
    env = NoopResetEnv(env, noop_max=30)
    env = MaxAndSkipEnv(env, skip=4)
    if sticky_actions:
        env = StickyActionEnv(env)
    if max_episode_steps is not None:
        env = TimeLimit(env, max_episode_steps=max_episode_steps)
    return env
